# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2022 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Base.Generator.dogboneII as dogboneII
import Path.Base.Language as PathLanguage
import Path.Dressup.DogboneII
import CAMTests.PathTestUtils as PathTestUtils
import math

PI = math.pi


class MockTB(object):
    def __init__(self, dia):
        self.Name = "ToolBit"
        self.Label = "ToolBit"
        self.Diameter = FreeCAD.Units.Quantity(dia, FreeCAD.Units.Length)


class MockTC(object):
    def __init__(self, dia=2):
        self.Name = "TC"
        self.Label = "TC"
        self.Tool = MockTB(dia)


class MockOp(object):
    def __init__(self, path, dia=2):
        self.Name = "OP"
        self.Label = "OP"
        self.Path = Path.Path(path)
        self.ToolController = MockTC(dia)


class MockFeaturePython(object):
    def __init__(self, name):
        self.prop = {}
        self.addProperty("App::PropertyString", "Name", val=name)
        self.addProperty("App::PropertyString", "Label", val=name)
        self.addProperty("App::PropertyLink", "Proxy")
        self.addProperty("Path::Path", "Path", val=Path.Path())

    def addProperty(self, typ, name, grp=None, desc=None, val=None):
        self.prop[name] = (typ, val)

    def setEditorMode(self, name, mode):
        pass

    def __setattr__(self, name, val):
        if name == "prop":
            return super().__setattr__(name, val)
        self.prop[name] = (self.prop[name][0], val)

    def __getattr__(self, name):
        if name == "prop":
            return super().__getattr__(name)
        typ, val = self.prop.get(name, (None, None))
        if typ is None and val is None:
            raise AttributeError
        if typ == "App::PropertyLength":
            if type(val) == float or type(val) == int:
                return FreeCAD.Units.Quantity(val, FreeCAD.Units.Length)
            return FreeCAD.Units.Quantity(val)
        return val


def CreateDressup(path):
    op = MockOp(path)
    obj = MockFeaturePython("DressupDogbone")
    db = Path.Dressup.DogboneII.Proxy(obj, op)
    obj.Proxy = db
    return obj


def MNVR(gcode, begin=None):
    # 'turns out the replace() isn't really necessary
    # leave it here anyway for clarity
    return PathLanguage.Maneuver.FromGCode(gcode.replace("/", "\n"), begin)


def INSTR(gcode, begin=None):
    return MNVR(gcode, begin).instr[0]


def KINK(gcode, begin=None):
    maneuver = MNVR(gcode, begin)
    if len(maneuver.instr) != 2:
        return None
    return dogboneII.Kink(maneuver.instr[0], maneuver.instr[1])


class TestDressupDogboneII(PathTestUtils.PathTestBase):
    """Unit tests for the DogboneII dressup."""

    def assertEqualPath(self, path, s):
        def cmd2str(cmd):
            param = [
                f"{k}{v:g}" if Path.Geom.isRoughly(0, v - int(v)) else f"{k}{v:.2f}"
                for k, v in cmd.Parameters.items()
            ]
            return f"{cmd.Name}{''.join(param)}"

        p = "/".join([cmd2str(cmd) for cmd in path.Commands])
        self.assertEqual(p, s)

    def test00(self):
        """Verify adaptive length"""

        def adaptive(k, a, n):
            return Path.Dressup.DogboneII.calc_length_adaptive(k, a, n, n)

        if True:
            # horizontal bones
            self.assertRoughly(adaptive(KINK("G1X1/G1X2"), 0, 1), 0)
            self.assertRoughly(adaptive(KINK("G1X1/G1Y1"), 0, 1), 1)
            self.assertRoughly(adaptive(KINK("G1X1/G1X2Y1"), 0, 1), 0.414214)
            self.assertRoughly(adaptive(KINK("G1X1/G1X0Y1"), 0, 1), 2.414211)
            self.assertRoughly(adaptive(KINK("G1X1/G1X0"), 0, 1), 1)
            self.assertRoughly(adaptive(KINK("G1X1/G1X0Y-1"), 0, 1), 2.414211)
            self.assertRoughly(adaptive(KINK("G1X1/G1X1Y-1"), 0, 1), 1)
            self.assertRoughly(adaptive(KINK("G1X1/G1X2Y-1"), 0, 1), 0.414214)
            self.assertRoughly(adaptive(KINK("G1X1Y1/G1X0Y2"), 0, 1), 0.414214)

        if True:
            # more horizontal and some vertical bones
            self.assertRoughly(adaptive(KINK("G1Y1/G1Y2"), 0, 1), 0)
            self.assertRoughly(adaptive(KINK("G1Y1/G1Y1X1"), PI, 1), 1)
            self.assertRoughly(adaptive(KINK("G1Y1/G1Y2X1"), PI, 1), 0.089820)
            self.assertRoughly(adaptive(KINK("G1Y1/G1Y2X1"), PI / 2, 1), 0.414214)
            self.assertRoughly(adaptive(KINK("G1Y1/G1Y0X1"), PI / 2, 1), 2.414211)
            self.assertRoughly(adaptive(KINK("G1Y1/G1Y0"), 0, 1), 1)
            self.assertRoughly(adaptive(KINK("G1Y1/G1Y0X-1"), PI / 2, 1), 2.414211)
            self.assertRoughly(adaptive(KINK("G1Y1/G1Y1X-1"), 0, 1), 1)
            self.assertRoughly(adaptive(KINK("G1Y1/G1Y2X-1"), 0, 1), 0.089820)
            self.assertRoughly(adaptive(KINK("G1Y1/G1Y2X-1"), PI / 2, 1), 0.414214)

        if True:
            # dogbones
            self.assertRoughly(adaptive(KINK("G1X1/G1Y1"), -PI / 4, 1), 0.414214)
            self.assertRoughly(adaptive(KINK("G1X1/G1X0Y1"), -PI / 8, 1), 1.613126)
            self.assertRoughly(adaptive(KINK("G1X1/G1Y-1"), PI / 4, 1), 0.414214)
            self.assertRoughly(adaptive(KINK("G1X1/G1X0Y-1"), PI / 8, 1), 1.613126)
            self.assertRoughly(adaptive(KINK("G1Y1/G1X-1"), PI / 4, 1), 0.414214)
            self.assertRoughly(adaptive(KINK("G1Y1/G1X1"), 3 * PI / 4, 1), 0.414214)
            self.assertRoughly(adaptive(KINK("G1Y-1/G1X1"), -3 * PI / 4, 1), 0.414214)
            self.assertRoughly(adaptive(KINK("G1Y-1/G1X-1"), -PI / 4, 1), 0.414214)
            self.assertRoughly(adaptive(KINK("G1X1Y1/G1X0Y2"), 0, 1), 0.414214)
            self.assertRoughly(adaptive(KINK("G1X-1Y1/G1X0Y2"), PI, 1), 0.414214)
            self.assertRoughly(adaptive(KINK("G1X1Y1/G1X2Y0"), PI / 2, 2), 0.828428)
            self.assertRoughly(adaptive(KINK("G1X-1Y-1/G1X-2Y0"), -PI / 2, 2), 0.828428)
            self.assertRoughly(adaptive(KINK("G1X-1Y1/G1X-2Y0"), PI / 2, 2), 0.828428)
            self.assertRoughly(adaptive(KINK("G1X1Y-1/G1X2Y0"), -PI / 2, 2), 0.828428)

    def test01(self):
        """Verify nominal length"""

        def nominal(k, a, n):
            return Path.Dressup.DogboneII.calc_length_nominal(k, a, n, 0)

        # neither angle nor kink matter
        self.assertRoughly(nominal(KINK("G1X1/G1X2"), 0, 13), 13)
        self.assertRoughly(nominal(KINK("G1X1/G1X2"), PI / 2, 13), 13)
        self.assertRoughly(nominal(KINK("G1X1/G1X2"), PI, 13), 13)
        self.assertRoughly(nominal(KINK("G1X1/G1X2"), -PI / 2, 13), 13)
        self.assertRoughly(nominal(KINK("G1X8/G1X12"), 0, 13), 13)
        self.assertRoughly(nominal(KINK("G1X9/G1X0"), 0, 13), 13)
        self.assertRoughly(nominal(KINK("G1X7/G1X9"), 0, 13), 13)
        self.assertRoughly(nominal(KINK("G1X5/G1X1"), 0, 13), 13)

    def test02(self):
        """Verify custom length"""

        def custom(k, a, c):
            return Path.Dressup.DogboneII.calc_length_custom(k, a, 0, c)

        # neither angle nor kink matter
        self.assertRoughly(custom(KINK("G1X1/G1X2"), 0, 7), 7)
        self.assertRoughly(custom(KINK("G1X1/G1X2"), PI / 2, 7), 7)
        self.assertRoughly(custom(KINK("G1X1/G1X2"), PI, 7), 7)
        self.assertRoughly(custom(KINK("G1X1/G1X2"), -PI / 2, 7), 7)
        self.assertRoughly(custom(KINK("G1X8/G1X12"), 0, 7), 7)
        self.assertRoughly(custom(KINK("G1X9/G1X0"), 0, 7), 7)
        self.assertRoughly(custom(KINK("G1X7/G1X9"), 0, 7), 7)
        self.assertRoughly(custom(KINK("G1X5/G1X1"), 0, 7), 7)

    def test10(self):
        """Verify basic op dressup"""

        obj = CreateDressup("G1X10/G1Y20")
        obj.Incision = Path.Dressup.DogboneII.Incision.Fixed
        obj.Style = Path.Dressup.DogboneII.Style.Tbone_H

        # bones on right side
        obj.Side = Path.Dressup.DogboneII.Side.Right
        obj.Proxy.execute(obj)
        self.assertEqualPath(obj.Path, "G1X10/G1X11/G1X10/G1Y20")

        # no bones on left side
        obj.Side = Path.Dressup.DogboneII.Side.Left
        obj.Proxy.execute(obj)
        self.assertEqualPath(obj.Path, "G1X10/G1Y20")

    def test11(self):
        """Verify retaining non-move instructions"""

        obj = CreateDressup("G1X10/(some comment)/G1Y20")
        obj.Incision = Path.Dressup.DogboneII.Incision.Fixed
        obj.Style = Path.Dressup.DogboneII.Style.Tbone_H

        # bone on right side
        obj.Side = Path.Dressup.DogboneII.Side.Right
        obj.Proxy.execute(obj)
        self.assertEqualPath(obj.Path, "G1X10/(some comment)/G1X11/G1X10/G1Y20")

        # no bone on left side
        obj.Side = Path.Dressup.DogboneII.Side.Left
        obj.Proxy.execute(obj)
        self.assertEqualPath(obj.Path, "G1X10/(some comment)/G1Y20")

    def test20(self):
        """Verify bone on plunge moves"""

        obj = CreateDressup("G0Z10/G1Z0/G1X10/G1Y10/G1X0/G1Y0/G0Z10")
        obj.Incision = Path.Dressup.DogboneII.Incision.Fixed
        obj.Style = Path.Dressup.DogboneII.Style.Tbone_H
        obj.Side = Path.Dressup.DogboneII.Side.Right

        obj.Proxy.execute(obj)
        self.assertEqualPath(
            obj.Path,
            "G0Z10/G1Z0/G1X10/G1X11/G1X10/G1Y10/G1X11/G1X10/G1X0/G1X-1/G1X0/G1Y0/G1X-1/G1X0/G0Z10",
        )

    def test21(self):
        """Verify ignoring plunge moves that don't connect"""

        obj = CreateDressup("G0Z10/G1Z0/G1X10/G1Y10/G1X0/G1Y5/G0Z10")
        obj.Incision = Path.Dressup.DogboneII.Incision.Fixed
        obj.Style = Path.Dressup.DogboneII.Style.Tbone_H
        obj.Side = Path.Dressup.DogboneII.Side.Right

        obj.Proxy.execute(obj)
        self.assertEqualPath(
            obj.Path,
            "G0Z10/G1Z0/G1X10/G1X11/G1X10/G1Y10/G1X11/G1X10/G1X0/G1X-1/G1X0/G1Y5/G0Z10",
        )

    def test30(self):
        """Verify TBone_V style"""

        def check_tbone(d, i, path, out, right):
            obj = CreateDressup(f"({d}.{i:02})/{path}")
            obj.Incision = Path.Dressup.DogboneII.Incision.Fixed
            if right:
                obj.Side = Path.Dressup.DogboneII.Side.Right
            else:
                obj.Side = Path.Dressup.DogboneII.Side.Left
            obj.Style = Path.Dressup.DogboneII.Style.Tbone_V
            obj.Proxy.execute(obj)
            self.assertEqualPath(obj.Path, f"({d}.{i:02})/{out}")

        # test data with a horizontal lead in
        test_data_h = [
            # top right quadrant
            ("G1X10Y0/G1X10Y10", "G1X10Y0/G1Y-1/G1Y0/G1X10Y10", True),
            ("G1X10Y0/G1X20Y10", "G1X10Y0/G1Y-1/G1Y0/G1X20Y10", True),
            ("G1X10Y0/G1X90Y10", "G1X10Y0/G1Y-1/G1Y0/G1X90Y10", True),
            ("G1X10Y0/G1X0Y10", "G1X10Y0/G1Y-1/G1Y0/G1X0Y10", True),
            # bottom right quadrant
            ("G1X10Y0/G1X90Y-10", "G1X10Y0/G1Y1/G1Y0/G1X90Y-10", False),
            ("G1X10Y0/G1X20Y-10", "G1X10Y0/G1Y1/G1Y0/G1X20Y-10", False),
            ("G1X10Y0/G1X10Y-10", "G1X10Y0/G1Y1/G1Y0/G1X10Y-10", False),
            ("G1X10Y0/G1X0Y-10", "G1X10Y0/G1Y1/G1Y0/G1X0Y-10", False),
            # top left quadrant
            ("G1X-10Y0/G1X-10Y10", "G1X-10Y0/G1Y-1/G1Y0/G1X-10Y10", False),
            ("G1X-10Y0/G1X-20Y10", "G1X-10Y0/G1Y-1/G1Y0/G1X-20Y10", False),
            ("G1X-10Y0/G1X-90Y10", "G1X-10Y0/G1Y-1/G1Y0/G1X-90Y10", False),
            ("G1X-10Y0/G1X-0Y10", "G1X-10Y0/G1Y-1/G1Y0/G1X-0Y10", False),
            # bottom left quadrant
            ("G1X-10Y0/G1X-90Y-10", "G1X-10Y0/G1Y1/G1Y0/G1X-90Y-10", True),
            ("G1X-10Y0/G1X-20Y-10", "G1X-10Y0/G1Y1/G1Y0/G1X-20Y-10", True),
            ("G1X-10Y0/G1X-10Y-10", "G1X-10Y0/G1Y1/G1Y0/G1X-10Y-10", True),
            ("G1X-10Y0/G1X-0Y-10", "G1X-10Y0/G1Y1/G1Y0/G1X-0Y-10", True),
        ]

        for i, (path, out, right) in enumerate(test_data_h):
            check_tbone("h", i, path, out, right)

        # test data with a vertical lead in
        test_data_v = [
            # top right quadrant
            ("G1X0Y10/G1X10Y10", "G1X0Y10/G1Y11/G1Y10/G1X10Y10", False),
            ("G1X0Y10/G1X10Y20", "G1X0Y10/G1Y11/G1Y10/G1X10Y20", False),
            ("G1X0Y10/G1X10Y90", "G1X0Y10/G1Y11/G1Y10/G1X10Y90", False),
            ("G1X0Y10/G1X10Y0", "G1X0Y10/G1Y11/G1Y10/G1X10Y0", False),
            # bottom right quadrant
            ("G1X0Y-10/G1X10Y-90", "G1X0Y-10/G1Y-11/G1Y-10/G1X10Y-90", True),
            ("G1X0Y-10/G1X10Y-20", "G1X0Y-10/G1Y-11/G1Y-10/G1X10Y-20", True),
            ("G1X0Y-10/G1X10Y-10", "G1X0Y-10/G1Y-11/G1Y-10/G1X10Y-10", True),
            ("G1X0Y-10/G1X10Y-0", "G1X0Y-10/G1Y-11/G1Y-10/G1X10Y-0", True),
            # top left quadrant
            ("G1X0Y10/G1X-10Y10", "G1X0Y10/G1Y11/G1Y10/G1X-10Y10", True),
            ("G1X0Y10/G1X-10Y20", "G1X0Y10/G1Y11/G1Y10/G1X-10Y20", True),
            ("G1X0Y10/G1X-10Y90", "G1X0Y10/G1Y11/G1Y10/G1X-10Y90", True),
            ("G1X0Y10/G1X-10Y0", "G1X0Y10/G1Y11/G1Y10/G1X-10Y0", True),
            # bottom left quadrant
            ("G1X0Y-10/G1X-10Y-90", "G1X0Y-10/G1Y-11/G1Y-10/G1X-10Y-90", False),
            ("G1X0Y-10/G1X-10Y-20", "G1X0Y-10/G1Y-11/G1Y-10/G1X-10Y-20", False),
            ("G1X0Y-10/G1X-10Y-10", "G1X0Y-10/G1Y-11/G1Y-10/G1X-10Y-10", False),
            ("G1X0Y-10/G1X-10Y-0", "G1X0Y-10/G1Y-11/G1Y-10/G1X-10Y-0", False),
        ]

        for i, (path, out, right) in enumerate(test_data_v):
            check_tbone("v", i, path, out, right)

    def test40(self):
        """Verify TBone_S style"""

        def check_tbone_s(d, i, path, out, right):
            obj = CreateDressup(f"(m{d}.{i:02})/{path}")
            obj.Incision = Path.Dressup.DogboneII.Incision.Fixed
            if right:
                obj.Side = Path.Dressup.DogboneII.Side.Right
            else:
                obj.Side = Path.Dressup.DogboneII.Side.Left
            obj.Style = Path.Dressup.DogboneII.Style.Tbone_S
            obj.Proxy.execute(obj)
            self.assertEqualPath(obj.Path, f"(m{d}.{i:02})/{out}")

        # short edge m0
        test_data_0 = [
            # CCW
            ("G1X10/G1Y20", "G1X10/G1Y-1/G1Y0/G1Y20", True),
            ("G1X10Y10/G1X-10Y30", "G1X10Y10/G1X10.71Y9.29/G1X10Y10/G1X-10Y30", True),
            ("G1Y10/G1X-20", "G1Y10/G1X1/G1X0/G1X-20", True),
            (
                "G1X-10Y10/G1X-30Y-10",
                "G1X-10Y10/G1X-9.29Y10.71/G1X-10Y10/G1X-30Y-10",
                True,
            ),
            ("G1X-10/G1Y-20", "G1X-10/G1Y1/G1Y0/G1Y-20", True),
            (
                "G1X-10Y-10/G1X10Y-30",
                "G1X-10Y-10/G1X-10.71Y-9.29/G1X-10Y-10/G1X10Y-30",
                True,
            ),
            ("G1Y-10/G1X20", "G1Y-10/G1X-1/G1X0/G1X20", True),
            ("G1X10Y-10/G1X30Y10", "G1X10Y-10/G1X9.29Y-10.71/G1X10Y-10/G1X30Y10", True),
            # CW
            ("G1X10/G1Y-20", "G1X10/G1Y1/G1Y0/G1Y-20", False),
            ("G1X10Y10/G1X30Y-10", "G1X10Y10/G1X9.29Y10.71/G1X10Y10/G1X30Y-10", False),
            ("G1Y10/G1X20", "G1Y10/G1X-1/G1X0/G1X20", False),
            (
                "G1X-10Y10/G1X10Y30",
                "G1X-10Y10/G1X-10.71Y9.29/G1X-10Y10/G1X10Y30",
                False,
            ),
            ("G1X-10/G1Y20", "G1X-10/G1Y-1/G1Y0/G1Y20", False),
            (
                "G1X-10Y-10/G1X-30Y10",
                "G1X-10Y-10/G1X-9.29Y-10.71/G1X-10Y-10/G1X-30Y10",
                False,
            ),
            ("G1Y-10/G1X-20", "G1Y-10/G1X1/G1X0/G1X-20", False),
            (
                "G1X10Y-10/G1X-10Y-30",
                "G1X10Y-10/G1X10.71Y-9.29/G1X10Y-10/G1X-10Y-30",
                False,
            ),
        ]

        for i, (path, out, right) in enumerate(test_data_0):
            check_tbone_s("0", i, path, out, right)

        # short edge m1
        test_data_1 = [
            # CCW
            ("G1X20/G1Y10", "G1X20/G1X21/G1X20/G1Y10", True),
            ("G1X20Y20/G1X10Y30", "G1X20Y20/G1X20.71Y20.71/G1X20Y20/G1X10Y30", True),
            ("G1Y20/G1X-10", "G1Y20/G1Y21/G1Y20/G1X-10", True),
            (
                "G1X-20Y20/G1X-30Y10",
                "G1X-20Y20/G1X-20.71Y20.71/G1X-20Y20/G1X-30Y10",
                True,
            ),
            ("G1X-20/G1Y-10", "G1X-20/G1X-21/G1X-20/G1Y-10", True),
            (
                "G1X-20Y-20/G1X-10Y-30",
                "G1X-20Y-20/G1X-20.71Y-20.71/G1X-20Y-20/G1X-10Y-30",
                True,
            ),
            ("G1Y-20/G1X10", "G1Y-20/G1Y-21/G1Y-20/G1X10", True),
            (
                "G1X20Y-20/G1X30Y-10",
                "G1X20Y-20/G1X20.71Y-20.71/G1X20Y-20/G1X30Y-10",
                True,
            ),
            # CW
            ("G1X20/G1Y-10", "G1X20/G1X21/G1X20/G1Y-10", False),
            ("G1X20Y20/G1X30Y10", "G1X20Y20/G1X20.71Y20.71/G1X20Y20/G1X30Y10", False),
            ("G1Y20/G1X10", "G1Y20/G1Y21/G1Y20/G1X10", False),
            (
                "G1X-20Y20/G1X-10Y30",
                "G1X-20Y20/G1X-20.71Y20.71/G1X-20Y20/G1X-10Y30",
                False,
            ),
            ("G1X-20/G1Y10", "G1X-20/G1X-21/G1X-20/G1Y10", False),
            (
                "G1X-20Y-20/G1X-30Y-10",
                "G1X-20Y-20/G1X-20.71Y-20.71/G1X-20Y-20/G1X-30Y-10",
                False,
            ),
            ("G1Y-20/G1X-10", "G1Y-20/G1Y-21/G1Y-20/G1X-10", False),
            (
                "G1X20Y-20/G1X10Y-30",
                "G1X20Y-20/G1X20.71Y-20.71/G1X20Y-20/G1X10Y-30",
                False,
            ),
        ]

        for i, (path, out, right) in enumerate(test_data_1):
            check_tbone_s("1", i, path, out, right)

    def test50(self):
        """Verify TBone_L style"""

        def check_tbone_l(d, i, path, out, right):
            obj = CreateDressup(f"(m{d}.{i:02})/{path}")
            obj.Incision = Path.Dressup.DogboneII.Incision.Fixed
            if right:
                obj.Side = Path.Dressup.DogboneII.Side.Right
            else:
                obj.Side = Path.Dressup.DogboneII.Side.Left
            obj.Style = Path.Dressup.DogboneII.Style.Tbone_L
            obj.Proxy.execute(obj)
            self.assertEqualPath(obj.Path, f"(m{d}.{i:02})/{out}")

        # long edge m1
        test_data_1 = [
            # CCW
            ("G1X10/G1Y20", "G1X10/G1X11/G1X10/G1Y20", True),
            ("G1X10Y10/G1X-10Y30", "G1X10Y10/G1X10.71Y10.71/G1X10Y10/G1X-10Y30", True),
            ("G1Y10/G1X-20", "G1Y10/G1Y11/G1Y10/G1X-20", True),
            (
                "G1X-10Y10/G1X-30Y-10",
                "G1X-10Y10/G1X-10.71Y10.71/G1X-10Y10/G1X-30Y-10",
                True,
            ),
            ("G1X-10/G1Y-20", "G1X-10/G1X-11/G1X-10/G1Y-20", True),
            (
                "G1X-10Y-10/G1X10Y-30",
                "G1X-10Y-10/G1X-10.71Y-10.71/G1X-10Y-10/G1X10Y-30",
                True,
            ),
            ("G1Y-10/G1X20", "G1Y-10/G1Y-11/G1Y-10/G1X20", True),
            (
                "G1X10Y-10/G1X30Y10",
                "G1X10Y-10/G1X10.71Y-10.71/G1X10Y-10/G1X30Y10",
                True,
            ),
            # CW
            ("G1X10/G1Y-20", "G1X10/G1X11/G1X10/G1Y-20", False),
            ("G1X10Y10/G1X30Y-10", "G1X10Y10/G1X10.71Y10.71/G1X10Y10/G1X30Y-10", False),
            ("G1Y10/G1X20", "G1Y10/G1Y11/G1Y10/G1X20", False),
            (
                "G1X-10Y10/G1X10Y30",
                "G1X-10Y10/G1X-10.71Y10.71/G1X-10Y10/G1X10Y30",
                False,
            ),
            ("G1X-10/G1Y20", "G1X-10/G1X-11/G1X-10/G1Y20", False),
            (
                "G1X-10Y-10/G1X-30Y10",
                "G1X-10Y-10/G1X-10.71Y-10.71/G1X-10Y-10/G1X-30Y10",
                False,
            ),
            ("G1Y-10/G1X-20", "G1Y-10/G1Y-11/G1Y-10/G1X-20", False),
            (
                "G1X10Y-10/G1X-10Y-30",
                "G1X10Y-10/G1X10.71Y-10.71/G1X10Y-10/G1X-10Y-30",
                False,
            ),
        ]

        for i, (path, out, right) in enumerate(test_data_1):
            check_tbone_l("1", i, path, out, right)

        # long edge m0
        test_data_0 = [
            # CCW
            ("G1X20/G1Y10", "G1X20/G1Y-1/G1Y0/G1Y10", True),
            ("G1X20Y20/G1X10Y30", "G1X20Y20/G1X20.71Y19.29/G1X20Y20/G1X10Y30", True),
            ("G1Y20/G1X-10", "G1Y20/G1X1/G1X0/G1X-10", True),
            (
                "G1X-20Y20/G1X-30Y10",
                "G1X-20Y20/G1X-19.29Y20.71/G1X-20Y20/G1X-30Y10",
                True,
            ),
            ("G1X-20/G1Y-10", "G1X-20/G1Y1/G1Y0/G1Y-10", True),
            (
                "G1X-20Y-20/G1X-10Y-30",
                "G1X-20Y-20/G1X-20.71Y-19.29/G1X-20Y-20/G1X-10Y-30",
                True,
            ),
            ("G1Y-20/G1X10", "G1Y-20/G1X-1/G1X0/G1X10", True),
            (
                "G1X20Y-20/G1X30Y-10",
                "G1X20Y-20/G1X19.29Y-20.71/G1X20Y-20/G1X30Y-10",
                True,
            ),
            # CW
            ("G1X20/G1Y-10", "G1X20/G1Y1/G1Y0/G1Y-10", False),
            ("G1X20Y20/G1X30Y10", "G1X20Y20/G1X19.29Y20.71/G1X20Y20/G1X30Y10", False),
            ("G1Y20/G1X10", "G1Y20/G1X-1/G1X0/G1X10", False),
            (
                "G1X-20Y20/G1X-10Y30",
                "G1X-20Y20/G1X-20.71Y19.29/G1X-20Y20/G1X-10Y30",
                False,
            ),
            ("G1X-20/G1Y10", "G1X-20/G1Y-1/G1Y0/G1Y10", False),
            (
                "G1X-20Y-20/G1X-30Y-10",
                "G1X-20Y-20/G1X-19.29Y-20.71/G1X-20Y-20/G1X-30Y-10",
                False,
            ),
            ("G1Y-20/G1X-10", "G1Y-20/G1X1/G1X0/G1X-10", False),
            (
                "G1X20Y-20/G1X10Y-30",
                "G1X20Y-20/G1X20.71Y-19.29/G1X20Y-20/G1X10Y-30",
                False,
            ),
        ]

        for i, (path, out, right) in enumerate(test_data_0):
            check_tbone_l("0", i, path, out, right)

    def test60(self):
        """Verify Dogbone style"""

        obj = CreateDressup("G1X10/G1Y20")
        obj.Incision = Path.Dressup.DogboneII.Incision.Fixed
        obj.Side = Path.Dressup.DogboneII.Side.Right

        obj.Style = Path.Dressup.DogboneII.Style.Dogbone
        obj.Proxy.execute(obj)
        self.assertEqualPath(obj.Path, "G1X10/G1X10.71Y-0.71/G1X10Y0/G1Y20")

    def test70(self):
        """Verify custom length."""

        obj = CreateDressup("G0Z10/G1Z0/G1X10/G1Y10/G1X0/G1Y0/G0Z10")
        obj.Style = Path.Dressup.DogboneII.Style.Tbone_H
        obj.Side = Path.Dressup.DogboneII.Side.Right

        obj.Incision = Path.Dressup.DogboneII.Incision.Custom
        obj.Custom = 3
        obj.Proxy.execute(obj)
        self.assertEqualPath(
            obj.Path,
            "G0Z10/G1Z0/G1X10/G1X13/G1X10/G1Y10/G1X13/G1X10/G1X0/G1X-3/G1X0/G1Y0/G1X-3/G1X0/G0Z10",
        )

        obj.Custom = 2
        obj.Proxy.execute(obj)
        self.assertEqualPath(
            obj.Path,
            "G0Z10/G1Z0/G1X10/G1X12/G1X10/G1Y10/G1X12/G1X10/G1X0/G1X-2/G1X0/G1Y0/G1X-2/G1X0/G0Z10",
        )

    def test80(self):
        """Verify adaptive length."""

        obj = CreateDressup("G1X10/G1Y20")
        obj.Incision = Path.Dressup.DogboneII.Incision.Adaptive
        obj.Side = Path.Dressup.DogboneII.Side.Right

        obj.Style = Path.Dressup.DogboneII.Style.Dogbone
        obj.Proxy.execute(obj)
        self.assertEqualPath(obj.Path, "G1X10/G1X10.29Y-0.29/G1X10Y0/G1Y20")

    def test81(self):
        """Verify adaptive length II."""

        obj = CreateDressup("G1X10/G1X20Y20")
        obj.Incision = Path.Dressup.DogboneII.Incision.Adaptive
        obj.Side = Path.Dressup.DogboneII.Side.Right

        obj.Style = Path.Dressup.DogboneII.Style.Dogbone
        obj.Proxy.execute(obj)
        self.assertEqualPath(obj.Path, "G1X10/G1X10.09Y-0.15/G1X10Y0/G1X20Y20")

    def test90(self):
        """Verify dogbone blacklist"""

        obj = CreateDressup("G0Z10/G1Z0/G1X10/G1Y10/G1X0/G1Y0/G0Z10")
        obj.Incision = Path.Dressup.DogboneII.Incision.Fixed
        obj.Style = Path.Dressup.DogboneII.Style.Tbone_H
        obj.Side = Path.Dressup.DogboneII.Side.Right
        obj.BoneBlacklist = [0, 2]
        obj.Proxy.execute(obj)
        self.assertEqualPath(
            obj.Path, "G0Z10/G1Z0/G1X10/G1Y10/G1X11/G1X10/G1X0/G1Y0/G1X-1/G1X0/G0Z10"
        )
        return obj

    def test91(self):
        """Verify dogbone on dogbone"""

        obj = self.test90()

        obj2 = MockFeaturePython("DressupDogbone001")
        db2 = Path.Dressup.DogboneII.Proxy(obj2, obj)
        obj2.Proxy = db2
        obj2.Incision = Path.Dressup.DogboneII.Incision.Fixed
        obj2.Style = Path.Dressup.DogboneII.Style.Tbone_H
        obj2.Side = Path.Dressup.DogboneII.Side.Right
        obj2.BoneBlacklist = [1]
        obj2.Proxy.execute(obj2)
        self.assertEqualPath(
            obj2.Path,
            "G0Z10/G1Z0/G1X10/G1X11/G1X10/G1Y10/G1X11/G1X10/G1X0/G1X-1/G1X0/G1Y0/G1X-1/G1X0/G0Z10",
        )



#!/usr/bin/env python3

# Kuka export lib (c) Jürgen Riegel 2009 LGPL 2.1 or higher

import time

TeachPointFold = """
;FOLD LIN P4  Vel= 0.2 m/s CPDAT1 Tool[1] Base[0];%{PE}%R 5.4.27,%MKUKATPBASIS,%CMOVE,%VLIN,%P 1:LIN, 2:P4, 3:, 5:0.2, 7:CPDAT1
$BWDSTART = FALSE
LDAT_ACT=LCPDAT1
FDAT_ACT=FP4
BAS(#CP_PARAMS,0.2)
LIN XP4
;ENDFOLD
"""

TeachPointDat = """
DECL E6POS XP4={X -25.1844196,Y 1122.42603,Z 1158.07996,A -14.3267002,B 0.537901878,C 179.028305,S 6,T 59,E1 0.0,E2 0.0,E3 0.0,E4 0.0,E5 0.0,E6 0.0}
DECL FDAT FP4={TOOL_NO 1,BASE_NO 0,IPO_FRAME #BASE,POINT2[] " "}
DECL LDAT LCPDAT1={VEL 2.0,ACC 100.0,APO_DIST 100.0,APO_FAC 50.0,ORI_TYP #VAR}
"""

HeaderSrc = """&ACCESS RVP
&REL 1
&PARAM TEMPLATE = C:\KRC\Roboter\Template\ExpertVorgabe
&PARAM EDITMASK = *
"""


def ExportCompactSub(Rob, Trak, FileName):
    print(Rob, Trak, FileName)
    Traj = Trak.Trajectory
    # open the output file
    SrcFile = open(FileName, "w")
    # header
    SrcFile.write(HeaderSrc)
    # subroutine definition
    SrcFile.write("DEF " + Trak.Name + "( )\n\n")
    SrcFile.write(";- Kuka src file, generated by FreeCAD (https://www.freecad.org)\n")
    SrcFile.write(";- " + time.asctime() + "\n\n")
    # defining world and base
    SrcFile.write(";------------- definitions ------------\n")
    SrcFile.write("EXT BAS (BAS_COMMAND :IN,REAL :IN ) ;set base to World\n")
    SrcFile.write("BAS (#INITMOV,0 ) ;Initialicing the defaults for Vel and so on \n\n")

    SrcFile.write("\n;------------- main part ------------\n")

    for w in Traj.Waypoints:
        (X, Y, Z) = (w.Pos.Base.x, w.Pos.Base.x, w.Pos.Base.x)
        (A, B, C) = w.Pos.Rotation.toEuler()
        V = w.Velocity / 1000.0  # from mm/s to m/s
        SrcFile.write("$VEL.CP = %f ; m/s ; m/s \n" % V)
        SrcFile.write(
            "LIN {X %.3f,Y %.3f,Z %.3f,A %.3f,B %.3f,C %.3f} ; %s\n"
            % (w.Pos.Base.x, w.Pos.Base.y, w.Pos.Base.z, A, B, C, w.Name)
        )

    # end of subroutine
    SrcFile.write("\n;------------- end ------------\n")
    SrcFile.write("END \n\n")

    SrcFile.close()
    # open the output file
    # DatFile = open(FileName[:-4]+'.dat','w')
    # header
    # DatFile.write(HeaderSrc)
    # subroutine definition
    # DatFile.write("DEFDAT "+Trak.Name+" PUBLIC\n\n")
    # DatFile.write("ENDDAT\n")
    # DatFile.close()


def ExportFullSub(Rob, Trak, FileName):
    print(Trak, FileName)



# ***************************************************************************
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM Magnetodynamics Elmer writer"
__author__ = "Uwe Stöhr"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

from FreeCAD import Console
from FreeCAD import Units

from femtools import femutils
from .. import sifio
from .. import writer as general_writer


class MgDynwriter:

    def __init__(self, writer, solver):
        self.write = writer
        self.solver = solver

    def getMagnetodynamicSolver(self, equation):
        # output the equation parameters
        s = self.write.createNonlinearSolver(equation)
        if not equation.IsHarmonic:
            s["Equation"] = "MgDyn"
            s["Procedure"] = sifio.FileAttr("MagnetoDynamics/WhitneyAVSolver")
            s["Variable"] = "av"
        else:
            s["Equation"] = "MgDynHarmonic"
            s["Procedure"] = sifio.FileAttr("MagnetoDynamics/WhitneyAVHarmonicSolver")
            s["Variable"] = "av[av re:1 av im:1]"
            frequency = equation.AngularFrequency.getValueAs("Hz")
            s["Angular Frequency"] = frequency
        s["Exec Solver"] = "Always"
        s["Optimize Bandwidth"] = True
        s["Stabilize"] = equation.Stabilize
        if equation.LinearSystemRefactorize is True:
            s["Linear System Refactorize"] = True
        if equation.UsePiolaTransform is True:
            s["Use Piola Transform"] = True
        if equation.QuadraticApproximation is True:
            s["Quadratic Approximation"] = True
        if equation.StaticConductivity is True:
            s["Static Conductivity"] = True
        if equation.FixInputCurrentDensity is True:
            s["Fix Input Current Density"] = True
        if equation.AutomatedSourceProjectionBCs is True:
            s["Automated Source Projection BCs"] = True
        if equation.UseLagrangeGauge is True:
            s["Use Lagrange Gauge"] = True
        if equation.LagrangeGaugePenalizationCoefficient != 0.0:
            s["Lagrange Gauge Penalization Coefficient"] = (
                equation.LagrangeGaugePenalizationCoefficient
            )
        if equation.UseTreeGauge is True:
            s["Use Tree Gauge"] = True
        return s

    def getMagnetodynamicSolverPost(self, equation):
        # output the equation parameters
        s = self.write.createNonlinearSolver(equation)
        s["Equation"] = "MgDynPost"
        s["Exec Solver"] = "Before Saving"
        s["Procedure"] = sifio.FileAttr("MagnetoDynamics/MagnetoDynamicsCalcFields")
        if equation.IsHarmonic:
            frequency = equation.AngularFrequency.getValueAs("Hz")
            s["Angular Frequency"] = frequency
        s["Potential Variable"] = "av"
        if equation.CalculateCurrentDensity is True:
            s["Calculate Current Density"] = True
        if equation.CalculateElectricField is True:
            s["Calculate Electric Field"] = True
        if equation.CalculateElementalFields is False:
            s["Calculate Elemental Fields"] = False
        if equation.CalculateHarmonicLoss is True:
            s["Calculate Harmonic Loss"] = True
        if equation.CalculateJouleHeating is True:
            s["Calculate Joule Heating"] = True
        if equation.CalculateMagneticFieldStrength is True:
            s["Calculate Magnetic Field Strength"] = True
        if equation.CalculateMaxwellStress is True:
            s["Calculate Maxwell Stress"] = True
        if equation.CalculateNodalFields is False:
            s["Calculate Nodal Fields"] = False
        if equation.CalculateNodalForces is True:
            s["Calculate Nodal Forces"] = True
        if equation.CalculateNodalHeating is True:
            s["Calculate Nodal Heating"] = True
        if equation.DiscontinuousBodies is True:
            s["Discontinuous Bodies"] = True
        s["Optimize Bandwidth"] = True
        s["Stabilize"] = equation.Stabilize
        return s

    def handleMagnetodynamicConstants(self):
        permeability = Units.Quantity(self.write.constsdef["PermeabilityOfVacuum"]).getValueAs(
            "H/m"
        )
        self.write.constant("Permeability Of Vacuum", permeability)

        permittivity = Units.Quantity(self.write.constsdef["PermittivityOfVacuum"]).getValueAs(
            "F/m"
        )
        self.write.constant("Permittivity Of Vacuum", permittivity)

    def handleMagnetodynamicMaterial(self, bodies):
        # check that all bodies have a set material
        for name in bodies:
            if self.write.getBodyMaterial(name) is None:
                raise general_writer.WriteError(
                    f"The body {name} is not referenced in any material.\n\n"
                )
        for obj in self.write.getMember("App::MaterialObject"):
            m = obj.Material
            refs = obj.References[0][1] if obj.References else self.write.getAllBodies()
            for name in (n for n in refs if n in bodies):
                if "ElectricalConductivity" not in m:
                    Console.PrintMessage(f"m: {m}\n")
                    raise general_writer.WriteError(
                        "The electrical conductivity must be specified for all materials.\n\n"
                    )
                if "RelativePermeability" not in m:
                    Console.PrintMessage(f"m: {m}\n")
                    raise general_writer.WriteError(
                        "The relative permeability must be specified for all materials.\n\n"
                    )
                self.write.material(name, "Name", m["Name"])
                conductivity = Units.Quantity(m["ElectricalConductivity"]).getValueAs("S/m")
                self.write.material(name, "Electric Conductivity", conductivity)
                self.write.material(name, "Relative Permeability", float(m["RelativePermeability"]))
                # permittivity might be necessary for the post processor
                if "RelativePermittivity" in m:
                    self.write.material(
                        name, "Relative Permittivity", float(m["RelativePermittivity"])
                    )

    def _outputMagnetodynamicBodyForce(self, obj, name, equation):
        if femutils.is_derived_from(obj, "Fem::ConstraintCurrentDensity") and obj.Mode == "Custom":
            # output only if current density is enabled and needed
            if obj.EnableCurrentDensity_re_1:
                current_density = obj.CurrentDensity_re_1.getValueAs("A/m^2")
                self.write.bodyForce(name, "Current Density 1", current_density)
            if obj.EnableCurrentDensity_re_2:
                current_density = obj.CurrentDensity_re_2.getValueAs("A/m^2")
                self.write.bodyForce(name, "Current Density 2", current_density)
            if obj.EnableCurrentDensity_re_3:
                current_density = obj.CurrentDensity_re_3.getValueAs("A/m^2")
                self.write.bodyForce(name, "Current Density 3", current_density)
            # imaginaries are only needed for harmonic equation
            if equation.IsHarmonic:
                if obj.EnableCurrentDensity_im_1:
                    current_density = obj.CurrentDensity_im_1.getValueAs("A/m^2")
                    self.write.bodyForce(name, "Current Density Im 1", current_density)
                if obj.EnableCurrentDensity_im_2:
                    current_density = obj.CurrentDensity_im_2.getValueAs("A/m^2")
                    self.write.bodyForce(name, "Current Density Im 2", current_density)
                if obj.EnableCurrentDensity_im_3:
                    current_density = obj.CurrentDensity_im_3.getValueAs("A/m^2")
                    self.write.bodyForce(name, "Current Density Im 3", current_density)

        if femutils.is_derived_from(obj, "Fem::ConstraintMagnetization"):
            # output only if magnetization is enabled and needed
            if obj.EnableMagnetization_1:
                magnetization = obj.Magnetization_re_1.getValueAs("A/m")
                self.write.bodyForce(name, "Magnetization 1", magnetization)
            if obj.EnableMagnetization_2:
                magnetization = obj.Magnetization_re_2.getValueAs("A/m")
                self.write.bodyForce(name, "Magnetization 2", magnetization)
            if obj.EnableMagnetization_3:
                magnetization = obj.Magnetization_re_3.getValueAs("A/m")
                self.write.bodyForce(name, "Magnetization 3", magnetization)
            # imaginaries are only needed for harmonic equation
            if equation.IsHarmonic:
                if obj.EnableMagnetization_1:
                    magnetization = obj.Magnetization_im_1.getValueAs("A/m")
                    self.write.bodyForce(name, "Magnetization Im 1", magnetization)
                if obj.EnableMagnetization_2:
                    magnetization = obj.Magnetization_im_2.getValueAs("A/m")
                    self.write.bodyForce(name, "Magnetization Im 2", magnetization)
                if obj.EnableMagnetization_3:
                    magnetization = obj.Magnetization_im_3.getValueAs("A/m")
                    self.write.bodyForce(name, "Magnetization Im 3", magnetization)

        if femutils.is_derived_from(obj, "Fem::ConstraintElectrostaticPotential"):
            if obj.PotentialEnabled:
                # output only if potential is enabled and needed
                potential = obj.Potential.getValueAs("V")
                self.write.bodyForce(name, "Electric Potential", potential)
                # imaginary is only needed for harmonic equation
                if equation.IsHarmonic:
                    potential = obj.AV_im.getValueAs("V")
                    self.write.bodyForce(name, "Electric Potential Im", potential)

    def handleMagnetodynamicBodyForces(self, bodies, equation):
        # the current density can either be a body force or a boundary constraint
        # therefore only output here if a solid is referenced
        currentDensities = self.write.getMember("Fem::ConstraintCurrentDensity")
        for obj in currentDensities:
            if obj.References:
                firstName = obj.References[0][1][0]
                firstName = firstName.rstrip("0123456789")
                if firstName == "Solid":
                    for name in obj.References[0][1]:
                        self._outputMagnetodynamicBodyForce(obj, name, equation)
                    self.write.handled(obj)
            else:
                # if there is only one current density without a reference,
                # add it to all bodies
                if len(currentDensities) == 1:
                    for name in bodies:
                        self._outputMagnetodynamicBodyForce(obj, name, equation)
                else:
                    raise general_writer.WriteError(
                        "Several current density constraints found without reference to a body.\n"
                        "Please set a body for each current density constraint."
                    )
            self.write.handled(obj)

        magnetizations = self.write.getMember("Fem::ConstraintMagnetization")
        for obj in magnetizations:
            if obj.References:
                for name in obj.References[0][1]:
                    self._outputMagnetodynamicBodyForce(obj, name, equation)
                self.write.handled(obj)
            else:
                # if there is only one magnetization without a reference,
                # add it to all bodies
                if len(magnetizations) == 1:
                    for name in bodies:
                        self._outputMagnetodynamicBodyForce(obj, name, equation)
                else:
                    raise general_writer.WriteError(
                        "Several magnetization constraints found without reference to a body.\n"
                        "Please set a body for each current density constraint."
                    )
            self.write.handled(obj)

        # the potential can either be a body force or a boundary constraint
        # therefore only output here if a solid is referenced
        potentials = self.write.getMember("Fem::ConstraintElectrostaticPotential")
        for obj in potentials:
            if obj.References:
                firstName = obj.References[0][1][0]
                firstName = firstName.rstrip("0123456789")
                if firstName == "Solid":
                    for name in obj.References[0][1]:
                        # output only if potentiual is enabled and needed
                        self._outputMagnetodynamicBodyForce(obj, name, equation)
                    self.write.handled(obj)

    def _outputMagnetodynamicBndConditions(self, obj, name, equation):
        if femutils.is_derived_from(obj, "Fem::ConstraintCurrentDensity") and obj.Mode == "Normal":
            current_density = obj.NormalCurrentDensity_re.getValueAs("A/m^2")
            self.write.boundary(name, "Electric Current Density", current_density)
            # imaginaries are only needed for harmonic equation
            if equation.IsHarmonic:
                current_density = obj.NormalCurrentDensity_im.getValueAs("A/m^2")
                self.write.boundary(name, "Electric Current Density Im", current_density)

        if femutils.is_derived_from(obj, "Fem::ConstraintElectrostaticPotential"):
            if obj.EnableAV:
                potential = obj.AV_re.getValueAs("V")
                if equation.IsHarmonic:
                    self.write.boundary(name, "AV re", potential)
                    potential = obj.AV_im.getValueAs("V")
                    self.write.boundary(name, "AV im", potential)
                else:
                    self.write.boundary(name, "AV", potential)
            if obj.EnableAV_1:
                potential = obj.AV_re_1.getValueAs("Wb/m")
                if equation.IsHarmonic:
                    self.write.boundary(name, "AV re {e} 1", potential)
                    potential = obj.AV_im_1.getValueAs("Wb/m")
                    self.write.boundary(name, "AV im {e} 1", potential)
                else:
                    self.write.boundary(name, "AV {e} 1", potential)
            if obj.EnableAV_2:
                potential = obj.AV_re_2.getValueAs("Wb/m")
                if equation.IsHarmonic:
                    self.write.boundary(name, "AV re {e} 2", potential)
                    potential = obj.AV_im_2.getValueAs("Wb/m")
                    self.write.boundary(name, "AV im {e} 2", potential)
                else:
                    self.write.boundary(name, "AV {e} 2", potential)
            if obj.EnableAV_3:
                potential = obj.AV_re_3.getValueAs("Wb/m")
                if equation.IsHarmonic:
                    self.write.boundary(name, "AV re {e} 3", potential)
                    potential = obj.AV_im_3.getValueAs("Wb/m")
                    self.write.boundary(name, "AV im {e} 3", potential)
                else:
                    self.write.boundary(name, "AV {e} 3", potential)

    def handleMagnetodynamicBndConditions(self, equation):
        # the current density can either be a body force or a boundary constraint
        # therefore only output here if a face is referenced
        currentDensities = self.write.getMember("Fem::ConstraintCurrentDensity")
        for obj in currentDensities:
            if obj.References:
                firstName = obj.References[0][1][0]
                firstName = firstName.rstrip("0123456789")
                if firstName == "Face":
                    for name in obj.References[0][1]:
                        self._outputMagnetodynamicBndConditions(obj, name, equation)
                    self.write.handled(obj)

        # the potential can either be a body force or a boundary constraint
        # therefore only output here if a face is referenced
        potentials = self.write.getMember("Fem::ConstraintElectrostaticPotential")
        for obj in potentials:
            if obj.References:
                firstName = obj.References[0][1][0]
                firstName = firstName.rstrip("0123456789")
                if firstName == "Face":
                    for name in obj.References[0][1]:
                        # output the FreeCAD label as comment
                        if obj.Label:
                            self.write.boundary(name, "! FreeCAD Name", obj.Label)
                        # output only if potential is enabled and needed
                        self._outputMagnetodynamicBndConditions(obj, name, equation)
                    self.write.handled(obj)


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM mesh netgen document object"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

## @package mesh_netgen
#  \ingroup FEM
#  \brief mesh gmsh object

from FreeCAD import Base
from . import base_fempythonobject

_PropHelper = base_fempythonobject._PropHelper


class MeshNetgen(base_fempythonobject.BaseFemPythonObject):
    """
    A Fem::FemMeshShapeBaseObject python type, add Netgen specific properties
    """

    Type = "Fem::FemMeshNetgen"

    def __init__(self, obj):
        super().__init__(obj)

        for prop in self._get_properties():
            prop.add_to_object(obj)

    def _get_properties(self):
        prop = []

        prop.append(
            _PropHelper(
                type="App::PropertyLinkList",
                name="MeshRegionList",
                group="Base",
                doc="Refinements of the mesh",
                value=[],
            )
        )
        # mesh parameters
        prop.append(
            _PropHelper(
                type="App::PropertyString",
                name="Optimize3d",
                group="Mesh Parameters",
                doc="3d optimization strategy.\n"
                + "m: move nodes, M: move nodes, cheap functional\n"
                + "s: swap faces, c: combine elements, d: divide elements,\n"
                + "D: divide and join opposite edges, remove element,\n"
                + "p: plot, no pause, P: plot, Pause,\n"
                + "h: Histogramm, no pause, H: Histogramm, pause",
                value="cmdDmustm",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyIntegerConstraint",
                name="OptimizationSteps3d",
                group="Mesh Parameters",
                doc="Number of 3d optimization steps",
                value={"value": 3, "min": 0},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyString",
                name="Optimize2d",
                group="Mesh Parameters",
                doc="2d optimization strategy.\n"
                + "s: swap opt 6 lines/node, S: swap optimal elements,\n"
                + "m: move nodes, p: plot, no pause\n"
                + "P: plot, pause, c: combine",
                value="smcmSmcmSmcm",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyIntegerConstraint",
                name="OptimizationSteps2d",
                group="Mesh Parameters",
                doc="Number of 2d optimization steps",
                value={"value": 3, "min": 0},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyFloat",
                name="OptimizationErrorPower",
                group="Mesh Parameters",
                doc="Power of error to approximate max error optimization",
                value=2.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="BlockFill",
                group="Mesh Parameters",
                doc="Do block filling",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="FillDistance",
                group="Mesh Parameters",
                doc="Block filling up to distance",
                value=0.1,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyFloatConstraint",
                name="Safety",
                group="Mesh Parameters",
                doc="Radius of local environment (times h)",
                value={"value": 5.0, "min": 0, "step": 0.1},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyFloatConstraint",
                name="RelinnerSafety",
                group="Mesh Parameters",
                doc="Radius of active environment (times h)",
                value={"value": 3.0, "min": 0, "step": 0.1},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="LocalH",
                group="Mesh Parameters",
                doc="Use local h",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="UseLocalH",
                group="Mesh Parameters",
                doc="Use local H",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyFloatConstraint",
                name="GrowthRate",
                group="Mesh Parameters",
                doc="Grading for local h",
                value={"value": 0.3, "min": 0, "step": 0.1},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="Delaunay",
                group="Mesh Parameters",
                doc="Use Delaunay for 3d meshing",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="Delaunay2d",
                group="Mesh Parameters",
                doc="Use Delaunay for 2d meshing",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="MaxSize",
                group="Mesh Parameters",
                doc="Maximal mesh size",
                value="1000 mm",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="MinSize",
                group="Mesh Parameters",
                doc="Minimal mesh size",
                value="0 mm",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyFloatConstraint",
                name="CloseEdgeFactor",
                group="Mesh Parameters",
                doc="Factor to restrict meshing based on close edges",
                value={"value": 2.0, "min": 0, "step": 0.1},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="StartInSurface",
                group="Mesh Parameters",
                doc="Start surface meshing from everywhere in surface",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="CheckOverlap",
                group="Mesh Parameters",
                doc="Check overlapping surfaces",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="CheckOverlappingBoundary",
                group="Mesh Parameters",
                doc="Check overlapping surface mesh before volume meshing",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="CheckChartBoundary",
                group="Mesh Parameters",
                doc="Check chart boundary",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyFloatConstraint",
                name="CurvatureSafety",
                group="Mesh Parameters",
                doc="Safety factor for curvatures (elements per radius)",
                value={"value": 2.0, "min": 0, "step": 0.1},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyFloatConstraint",
                name="SegmentsPerEdge",
                group="Mesh Parameters",
                doc="Minimal number of segments per edge",
                value={"value": 2.0, "min": 0, "step": 0.1},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyFloatConstraint",
                name="ElementSizeWeight",
                group="Mesh Parameters",
                doc="Weight of element size respect to element shape",
                value={"value": 0.2, "min": 0, "step": 0.1},
            )
        )

        # Netgen meshing steps
        meshing_step = [
            "AnalyzeGeometry",
            "MeshEdges",
            "MeshSurface",
            "OptimizeSurface",
            "MeshVolume",
            "OptimizeVolume",
        ]

        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="StartStep",
                group="Mesh Parameters",
                doc="First step",
                value=meshing_step,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="EndStep",
                group="Mesh Parameters",
                doc="Last step",
                value=meshing_step,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyIntegerConstraint",
                name="GiveUpTolerance2d",
                group="Mesh Parameters",
                doc="Give up quality class, 2d meshing",
                value={"value": 200, "min": 0},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyIntegerConstraint",
                name="GiveUpTolerance",
                group="Mesh Parameters",
                doc="Give up quality class, 3d meshing",
                value={"value": 10, "min": 0},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyIntegerConstraint",
                name="GiveUpToleranceOpenQuads",
                group="Mesh Parameters",
                doc="Give up quality class, for closing open quads, greater than 100 for free pyramids",
                value={"value": 15, "min": 0},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyIntegerConstraint",
                name="MaxOuterSteps",
                group="Mesh Parameters",
                doc="Maximal outer steps",
                value={"value": 10, "min": 0},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyIntegerConstraint",
                name="StarShapeClass",
                group="Mesh Parameters",
                doc="Class starting star-shape filling",
                value={"value": 5, "min": 0},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyInteger",
                name="BaseElementNp",
                group="Mesh Parameters",
                doc="If non-zero, baseelement must have BaseElementlNp points",
                value=0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyIntegerConstraint",
                name="Sloppy",
                group="Mesh Parameters",
                doc="Quality tolerances are handled less careful",
                value={"value": 10, "min": 0},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyFloatConstraint",
                name="BadElementLimit",
                group="Mesh Parameters",
                doc="Limit for max element angle (150-180)",
                value={"value": 175, "min": 0},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="CheckImpossible",
                group="Mesh Parameters",
                doc="",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyInteger",
                name="Only3dDomainNr",
                group="Mesh Parameters",
                doc="",
                value=0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="SecondOrder",
                group="Mesh Parameters",
                doc="Second order element meshing",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="SecondOrderLinear",
                group="Mesh Parameters",
                doc="Second order nodes are created by linear interpolation",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyIntegerConstraint",
                name="ElementOrder",
                group="Mesh Parameters",
                doc="High order element curvature",
                value={"value": 1, "min": 1},
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="QuadDominated",
                group="Mesh Parameters",
                doc="Quad-dominated surface meshing",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="TryHexes",
                group="Mesh Parameters",
                doc="Try hexahedral elements",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="InvertTets",
                group="Mesh Parameters",
                doc="",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="InvertTrigs",
                group="Mesh Parameters",
                doc="",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="ZRefine",
                group="Mesh Parameters",
                doc="Z-refinement for extruded shapes",
                value=["No", "Regular", "Custom"],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyVector",
                name="ZRefineDirection",
                group="Mesh Parameters",
                doc="Z-refinement direction",
                value=Base.Vector(0, 0, 1),
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyFloatList",
                name="ZRefineSize",
                group="Mesh Parameters",
                doc="Z-refinement size given as a fraction of the shape size.\n"
                + "For a regular partition only one value is needed",
                value=[
                    0.1,
                ],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="ParallelMeshing",
                group="Mesh Parameters",
                doc="Use parallel meshing",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="HealShape",
                group="Mesh Parameters",
                doc="Heal shape before meshing",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="Glue",
                group="Mesh Parameters",
                doc="Glue shapes to get a conformal mesh",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="Fineness",
                group="Mesh Parameters",
                doc="Mesh granularity.\n"
                + "If differs from UserDefined, uses specific values\n"
                + "for CurvatureSafety, SegmentsPerEdge, GrowthRate,\n"
                + "CloseEdgeFactor and OptimizationSteps3d",
                value=["VeryCoarse", "Coarse", "Moderate", "Fine", "VeryFine", "UserDefined"],
            )
        )

        return prop

    def onChanged(self, obj, prop):
        if prop == "Fineness":
            if obj.Fineness != "UserDefined":
                p = self.get_predef_fineness_params(obj.Fineness)
                obj.CurvatureSafety = p["curvaturesafety"]
                obj.SegmentsPerEdge = p["segmentsperedge"]
                obj.GrowthRate = p["grading"]
                obj.CloseEdgeFactor = p["closeedgefac"]
                obj.OptimizationSteps3d = p["optsteps3d"]

                obj.setPropertyStatus("CurvatureSafety", "ReadOnly")
                obj.setPropertyStatus("SegmentsPerEdge", "ReadOnly")
                obj.setPropertyStatus("GrowthRate", "ReadOnly")
                obj.setPropertyStatus("CloseEdgeFactor", "ReadOnly")
                obj.setPropertyStatus("OptimizationSteps3d", "ReadOnly")
            else:
                obj.setPropertyStatus("CurvatureSafety", "-ReadOnly")
                obj.setPropertyStatus("SegmentsPerEdge", "-ReadOnly")
                obj.setPropertyStatus("GrowthRate", "-ReadOnly")
                obj.setPropertyStatus("CloseEdgeFactor", "-ReadOnly")
                obj.setPropertyStatus("OptimizationSteps3d", "-ReadOnly")

    def onDocumentRestored(self, obj):
        # update old project with new properties
        for prop in self._get_properties():
            try:
                obj.getPropertyByName(prop.name)
            except Base.PropertyError:
                prop.add_to_object(obj)

            # for StartStep and EndStep set enumeration index from old integer value
            if prop.name == "StartStep" or prop.name == "EndStep":
                prop.handle_change_type(
                    obj, "App::PropertyInteger", lambda x: 0 if x <= 1 else 5 if x >= 6 else x - 1
                )
                # update enum values
                setattr(obj, prop.name, prop.value)

    def get_predef_fineness_params(self, fineness):
        # set specific parameters by fineness
        params = {}
        if fineness == "VeryCoarse":
            params["curvaturesafety"] = 1
            params["segmentsperedge"] = 0.3
            params["grading"] = 0.7
            params["closeedgefac"] = 0.5
            params["optsteps3d"] = 5
        elif fineness == "Coarse":
            params["curvaturesafety"] = 1.5
            params["segmentsperedge"] = 0.5
            params["grading"] = 0.5
            params["closeedgefac"] = 1
            params["optsteps3d"] = 5
        elif fineness == "Moderate":
            params["curvaturesafety"] = 2
            params["segmentsperedge"] = 1
            params["grading"] = 0.3
            params["closeedgefac"] = 2
            params["optsteps3d"] = 5
        elif fineness == "Fine":
            params["curvaturesafety"] = 3
            params["segmentsperedge"] = 2
            params["grading"] = 0.2
            params["closeedgefac"] = 3.5
            params["optsteps3d"] = 5
        elif fineness == "VeryFine":
            params["curvaturesafety"] = 5
            params["segmentsperedge"] = 3
            params["grading"] = 0.1
            params["closeedgefac"] = 5
            params["optsteps3d"] = 5
        elif fineness == "UserDefined":
            params["curvaturesafety"] = 2
            params["segmentsperedge"] = 2
            params["grading"] = 0.3
            params["closeedgefac"] = 2
            params["optsteps3d"] = 3

        return params



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create LinearDimension or AngularDinemsion objects.

This includes linear dimensions, radial dimensions, and angular dimensions.
"""
## @package make_dimension
# \ingroup draftmake
# \brief Provides functions to create Linear or AngularDimension objects.

## \addtogroup draftmake
# @{
import math

import FreeCAD as App
import DraftVecUtils
import WorkingPlane

from draftgeoutils import edges
from draftutils import gui_utils
from draftutils import params
from draftutils import utils
from draftutils.messages import _wrn, _err
from draftutils.translate import translate

from draftobjects.dimension import LinearDimension, AngularDimension

if App.GuiUp:
    from draftviewproviders.view_dimension import (
        ViewProviderLinearDimension,
        ViewProviderAngularDimension,
    )


def _get_flip_text_lin(p1, p2, wp, normal):
    # for linear, radial dimensions
    if not params.get_param("DimAutoFlipText"):
        return False
    p1 = wp.project_point(p1)
    p2 = wp.project_point(p2)
    ang = DraftVecUtils.angle(wp.u, p2.sub(p1), normal)
    tol = 1e-4  # high tolerance
    if math.isclose(ang, 0, abs_tol=tol):
        return False
    if math.isclose(abs(ang), math.pi, abs_tol=tol):
        return True
    if math.isclose(ang, math.pi / 2, abs_tol=tol):
        return False
    if math.isclose(ang, -math.pi / 2, abs_tol=tol):
        return True
    # 90-180 (in that quadrant + 1st point closest to the origin):
    if math.pi / 2 < ang < math.pi:
        return True
    # 180-270:
    if -math.pi < ang < -math.pi / 2:
        return True
    # 0-90 and 270-360:
    return False


def _get_flip_text_ang(cen, sta, end, normal):
    # for angular dimensions
    if not params.get_param("DimAutoFlipText"):
        return False
    import Part

    circle = Part.makeCircle(1, cen, normal, sta, end)
    mid = edges.findMidpoint(circle)
    wp = WorkingPlane.get_working_plane(update=False)
    ang = DraftVecUtils.angle(wp.u, mid.sub(cen), normal)
    tol = 1e-4  # high tolerance
    if math.isclose(ang, 0, abs_tol=tol):
        return True
    if math.isclose(abs(ang), math.pi, abs_tol=tol):
        return False
    if ang > 0:
        return False
    return True


def make_dimension(p1, p2, p3=None, p4=None):
    """Create one of three types of dimension objects.

    In all dimensions the p3 parameter defines a point through which
    the dimension line will go through.

    The current line width and color will be used.

    Linear dimension
    ----------------
    - (p1, p2, p3): a simple linear dimension from p1 to p2

    - (object, i1, i2, p3): creates a linked dimension to the provided
      object (edge), measuring the distance between its vertices
      indexed i1 and i2

    Circular dimension
    ------------------
    - (arc, i1, mode, p3): creates a linked dimension to the given arc
      object, i1 is the index of the arc edge that will be measured;
      mode is either "radius" or "diameter".
    """
    if not App.ActiveDocument:
        _err("No active document. Aborting")
        return None

    new_obj = App.ActiveDocument.addObject("App::FeaturePython", "Dimension")
    LinearDimension(new_obj)

    wp = WorkingPlane.get_working_plane(update=False)
    normal = wp.axis
    flip_text = False
    override = ""

    if App.GuiUp:
        # invert the normal if we are viewing it from the back
        vnorm = gui_utils.get3DView().getViewDirection()
        if vnorm.getAngle(normal) < math.pi / 2:
            normal = normal.negative()

    new_obj.Normal = normal

    if isinstance(p1, App.Vector) and isinstance(p2, App.Vector):
        # Measure a straight distance between p1 and p2
        new_obj.Start = p1
        new_obj.End = p2
        if not p3:
            p3 = p2.sub(p1)
            p3.multiply(0.5)
            p3 = p1.add(p3)

        if App.GuiUp:
            flip_text = _get_flip_text_lin(p1, p2, wp, normal)

    elif isinstance(p2, int) and isinstance(p3, int):
        # p1 is an object, and measure the distance between vertices p2 and p3
        # of this object
        linked = []
        idx = (p2, p3)
        linked.append((p1, "Vertex" + str(p2 + 1)))
        linked.append((p1, "Vertex" + str(p3 + 1)))
        new_obj.LinkedGeometry = linked
        new_obj.Support = p1

        v1 = p1.Shape.Vertexes[idx[0]].Point
        v2 = p1.Shape.Vertexes[idx[1]].Point
        # p4, and now p3, is the point through which the dimension line will pass
        p3 = p4
        if not p3:
            p3 = v2.sub(v1)
            p3.multiply(0.5)
            p3 = v1.add(p3)

        if App.GuiUp:
            flip_text = _get_flip_text_lin(v1, v2, wp, normal)

    elif isinstance(p3, str):
        # If the original p3 is a string, we are measuring a circular arc
        # p2 should be an integer number starting from 0
        linked = []
        linked.append((p1, "Edge" + str(p2 + 1)))

        if p3 == "radius":
            # linked.append((p1, "Center"))
            override = "R $dim"
            new_obj.Diameter = False
        elif p3 == "diameter":
            # linked.append((p1, "Diameter"))
            override = "Ø $dim"
            new_obj.Diameter = True
        new_obj.LinkedGeometry = linked
        new_obj.Support = p1

        cen = p1.Shape.Edges[p2].Curve.Center
        # p4, and now p3, is the point through which the dimension line will pass
        p3 = p4
        if not p3:
            p3 = cen.add(App.Vector(1, 0, 0))

        if App.GuiUp:
            flip_text = _get_flip_text_lin(cen, p3, wp, normal)

    # This p3 is the point through which the dimension line will pass,
    # but this may not be the original p3, it could have been p4
    # depending on the first three parameter values
    new_obj.Dimline = p3

    if App.GuiUp:
        ViewProviderLinearDimension(new_obj.ViewObject)
        new_obj.ViewObject.FlipText = flip_text
        new_obj.ViewObject.Override = override
        gui_utils.format_object(new_obj)
        gui_utils.select(new_obj)

    return new_obj


def makeDimension(p1, p2, p3=None, p4=None):
    """Create a dimension. DEPRECATED. Use 'make_dimension'."""
    _wrn(translate("draft", "This function is deprecated. Do not use this function directly."))
    _wrn(translate("draft", "Use one of 'make_linear_dimension', or 'make_linear_dimension_obj'."))

    return make_dimension(p1, p2, p3, p4)


def make_linear_dimension(p1, p2, dim_line=None):
    """Create a free linear dimension from two main points.

    Parameters
    ----------
    p1: Base::Vector3
        First point of the measurement.

    p2: Base::Vector3
        Second point of the measurement.

    dim_line: Base::Vector3, optional
        It defaults to `None`.
        This is a point through which the extension of the dimension line
        will pass.
        This point controls how close or how far the dimension line is
        positioned from the measured segment that goes from `p1` to `p2`.

        If it is `None`, this point will be calculated from the intermediate
        distance between `p1` and `p2`.

    Returns
    -------
    App::FeaturePython
        A scripted object of type `'LinearDimension'`.
        This object does not have a `Shape` attribute, as the text and lines
        are created on screen by Coin (pivy).

    None
        If there is a problem it will return `None`.
    """
    _name = "make_linear_dimension"

    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return None

    try:
        utils.type_check([(p1, App.Vector)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a vector."))
        return None

    try:
        utils.type_check([(p2, App.Vector)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a vector."))
        return None

    if dim_line:
        try:
            utils.type_check([(dim_line, App.Vector)], name=_name)
        except TypeError:
            _err(translate("draft", "Wrong input: must be a vector."))
            return None
    else:
        diff = p2.sub(p1)
        diff.multiply(0.5)
        dim_line = p1.add(diff)

    new_obj = make_dimension(p1, p2, dim_line)

    return new_obj


def make_linear_dimension_obj(edge_object, i1=1, i2=2, dim_line=None):
    """Create a linear dimension from an object.

    Parameters
    ----------
    edge_object: Part::Feature
        The object which has an edge which will be measured.
        It must have a `Part::TopoShape`, and at least one element
        in `Shape.Vertexes`, to be able to measure a distance.

    i1: int, optional
        It defaults to `1`.
        It is the index of the first vertex in `edge_object` from which
        the measurement will be taken.
        The minimum value should be `1`, which will be interpreted
        as `'Vertex1'`.

        If the value is below `1`, it will be set to `1`.

    i2: int, optional
        It defaults to `2`, which will be converted to `'Vertex2'`.
        It is the index of the second vertex in `edge_object` that determines
        the endpoint of the measurement.

        If it is the same value as `i1`, the resulting measurement will be
        made from the origin `(0, 0, 0)` to the vertex indicated by `i1`.

        If the value is below `1`, it will be set to the last vertex
        in `edge_object`.

        Then to measure the first and last, this could be used
        ::
            make_linear_dimension_obj(edge_object, i1=1, i2=-1)

    dim_line: Base::Vector3
        It defaults to `None`.
        This is a point through which the extension of the dimension line
        will pass.
        This point controls how close or how far the dimension line is
        positioned from the measured segment in `edge_object`.

        If it is `None`, this point will be calculated from the intermediate
        distance between the vertices defined by `i1` and `i2`.

    Returns
    -------
    App::FeaturePython
        A scripted object of type `'LinearDimension'`.
        This object does not have a `Shape` attribute, as the text and lines
        are created on screen by Coin (pivy).

    None
        If there is a problem it will return `None`.
    """
    _name = "make_linear_dimension_obj"

    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return None

    if isinstance(edge_object, (list, tuple)):
        _err(translate("draft", "Wrong input: edge_object must not be a list or tuple."))
        return None

    found, edge_object = utils.find_object(edge_object, doc)
    if not found:
        _err(translate("draft", "Wrong input: edge_object not in document."))
        return None

    if not hasattr(edge_object, "Shape"):
        _err(translate("draft", "Wrong input: object doesn't have a 'Shape' to measure."))
        return None
    if not hasattr(edge_object.Shape, "Vertexes") or len(edge_object.Shape.Vertexes) < 1:
        _err(
            translate(
                "draft",
                "Wrong input: object does not have at least 1 element in 'Vertexes' to use for measuring.",
            )
        )
        return None

    try:
        utils.type_check([(i1, int)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be an integer."))
        return None

    if i1 < 1:
        i1 = 1
        _wrn(translate("draft", "i1: values below 1 are not allowed; will be set to 1."))

    vx1 = edge_object.getSubObject("Vertex" + str(i1))
    if not vx1:
        _err(translate("draft", "Wrong input: vertex not in object."))
        return None

    try:
        utils.type_check([(i2, int)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a vector."))
        return None

    if i2 < 1:
        i2 = len(edge_object.Shape.Vertexes)
        _wrn(
            translate(
                "draft",
                "i2: values below 1 are not allowed; will be set to the last vertex in the object.",
            )
        )

    vx2 = edge_object.getSubObject("Vertex" + str(i2))
    if not vx2:
        _err(translate("draft", "Wrong input: vertex not in object."))
        return None

    if dim_line:
        try:
            utils.type_check([(dim_line, App.Vector)], name=_name)
        except TypeError:
            _err(translate("draft", "Wrong input: must be a vector."))
            return None
    else:
        diff = vx2.Point.sub(vx1.Point)
        diff.multiply(0.5)
        dim_line = vx1.Point.add(diff)

    # TODO: the internal function expects an index starting with 0
    # so we need to decrease the value here.
    # This should be changed in the future in the internal function.
    i1 -= 1
    i2 -= 1

    new_obj = make_dimension(edge_object, i1, i2, dim_line)

    return new_obj


def make_radial_dimension_obj(edge_object, index=1, mode="radius", dim_line=None):
    """Create a radial or diameter dimension from an arc object.

    Parameters
    ----------
    edge_object: Part::Feature
        The object which has a circular edge which will be measured.
        It must have a `Part::TopoShape`, and at least one element
        must be a circular edge in `Shape.Edges` to be able to measure
        its radius.

    index: int, optional
        It defaults to `1`.
        It is the index of the edge in `edge_object` which is going to
        be measured.
        The minimum value should be `1`, which will be interpreted
        as `'Edge1'`. If the value is below `1`, it will be set to `1`.

    mode: str, optional
        It defaults to `'radius'`; the other option is `'diameter'`.
        It determines whether the dimension will be shown as a radius
        or as a diameter.

    dim_line: Base::Vector3, optional
        It defaults to `None`.
        This is a point through which the extension of the dimension line
        will pass. The dimension line will be a radius or diameter
        of the measured arc, extending from the center to the arc itself.

        If it is `None`, this point will be set to one unit to the right
        of the center of the arc, which will create a dimension line that is
        horizontal, that is, parallel to the +X axis.

    Returns
    -------
    App::FeaturePython
        A scripted object of type `'LinearDimension'`.
        This object does not have a `Shape` attribute, as the text and lines
        are created on screen by Coin (pivy).

    None
        If there is a problem it will return `None`.
    """
    _name = "make_radial_dimension_obj"

    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return None

    found, edge_object = utils.find_object(edge_object, doc)
    if not found:
        _err(translate("draft", "Wrong input: edge_object not in document."))
        return None

    if not hasattr(edge_object, "Shape"):
        _err(translate("draft", "Wrong input: object doesn't have a 'Shape' to measure."))
        return None
    if not hasattr(edge_object.Shape, "Edges") or len(edge_object.Shape.Edges) < 1:
        _err(
            translate(
                "draft",
                "Wrong input: object doesn't have at least one element in 'Edges' to use for measuring.",
            )
        )
        return None

    try:
        utils.type_check([(index, int)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be an integer."))
        return None

    if index < 1:
        index = 1
        _wrn(translate("draft", "index: values below 1 are not allowed; will be set to 1."))

    edge = edge_object.getSubObject("Edge" + str(index))
    if not edge:
        _err(translate("draft", "Wrong input: index doesn't correspond to an edge in the object."))
        return None

    if not hasattr(edge, "Curve") or edge.Curve.TypeId != "Part::GeomCircle":
        _err(translate("draft", "Wrong input: index doesn't correspond to a circular edge."))
        return None

    try:
        utils.type_check([(mode, str)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a string, 'radius' or 'diameter'."))
        return None

    if mode not in ("radius", "diameter"):
        _err(translate("draft", "Wrong input: must be a string, 'radius' or 'diameter'."))
        return None

    if dim_line:
        try:
            utils.type_check([(dim_line, App.Vector)], name=_name)
        except TypeError:
            _err(translate("draft", "Wrong input: must be a vector."))
            return None
    else:
        center = edge_object.Shape.Edges[index - 1].Curve.Center
        dim_line = center + App.Vector(1, 0, 0)

    # TODO: the internal function expects an index starting with 0
    # so we need to decrease the value here.
    # This should be changed in the future in the internal function.
    index -= 1

    new_obj = make_dimension(edge_object, index, mode, dim_line)

    return new_obj


def make_angular_dimension(
    center=App.Vector(0, 0, 0),
    angles=None,  # If None, set to [0,90]
    dim_line=App.Vector(10, 10, 0),
    normal=None,
):
    """Create an angular dimension from the given center and angles.

    Parameters
    ----------
    center: Base::Vector3, optional
        It defaults to the origin `Vector(0, 0, 0)`.
        Center of the dimension line, which is a circular arc.

    angles: list of two floats, optional
        It defaults to `[0, 90]`.
        It is a list of two angles, given in degrees, that determine
        the aperture of the dimension line, that is, of the circular arc.
        It is drawn counter-clockwise.
        ::
            angles = [0 90]
            angles = [330 60]  # the arc crosses the X axis
            angles = [-30 60]  # same angle

    dim_line: Base::Vector3, optional
        It defaults to `Vector(10, 10, 0)`.
        This is a point through which the extension of the dimension line
        will pass. This defines the radius of the dimension line,
        the circular arc.

    normal: Base::Vector3, optional
        It defaults to `None`, in which case the axis of the current working
        plane is used.

    Returns
    -------
    App::FeaturePython
        A scripted object of type `'AngularDimension'`.
        This object does not have a `Shape` attribute, as the text and lines
        are created on screen by Coin (pivy).

    None
        If there is a problem it will return `None`.
    """
    _name = "make_angular_dimension"

    # Prevent later modification of a default parameter by using a placeholder
    if angles is None:
        angles = [0, 90]

    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return None

    try:
        utils.type_check([(center, App.Vector)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a vector."))
        return None

    try:
        utils.type_check([(angles, (tuple, list))], name=_name)

        if len(angles) != 2:
            _err(translate("draft", "Wrong input: must be a list with two angles."))
            return None

        ang1, ang2 = angles
        utils.type_check([(ang1, (int, float)), (ang2, (int, float))], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a list with two angles."))
        return None

    # If the angle is larger than 360 degrees, make sure
    # it is smaller than 360
    for n in range(len(angles)):
        if angles[n] > 360:
            angles[n] = angles[n] - 360

    try:
        utils.type_check([(dim_line, App.Vector)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a vector."))
        return None

    if normal:
        try:
            utils.type_check([(dim_line, App.Vector)], name=_name)
        except TypeError:
            _err(translate("draft", "Wrong input: must be a vector."))
            return None

    if not normal:
        normal = WorkingPlane.get_working_plane(update=False).axis

    new_obj = App.ActiveDocument.addObject("App::FeaturePython", "Dimension")
    AngularDimension(new_obj)

    if App.GuiUp:
        ViewProviderAngularDimension(new_obj.ViewObject)

        # Invert the normal if we are viewing it from the back.
        # This is determined by the angle between the current
        # 3D view and the provided normal being below 90 degrees
        vnorm = gui_utils.get3DView().getViewDirection()
        if vnorm.getAngle(normal) < math.pi / 2:
            normal = normal.negative()

    new_obj.Center = center
    new_obj.FirstAngle = angles[0]
    new_obj.LastAngle = angles[1]
    new_obj.Dimline = dim_line
    new_obj.Normal = normal

    if App.GuiUp:
        new_obj.ViewObject.FlipText = _get_flip_text_ang(center, angles[0], angles[1], normal)
        gui_utils.format_object(new_obj)
        gui_utils.select(new_obj)

    return new_obj


def makeAngularDimension(center, angles, p3, normal=None):
    """Create an angle dimension. DEPRECATED. Use 'make_angular_dimension'."""
    utils.use_instead("make_angular_dimension")

    ang1, ang2 = angles
    angles = [math.degrees(ang2), math.degrees(ang1)]

    return make_angular_dimension(center=center, angles=angles, dim_line=p3, normal=normal)


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2024 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Defines a QWidget-derived class for displaying a set of buttons that affect the Addon
Manager as a whole (rather than a specific Addon). Typically inserted at the bottom of the Addon
Manager main window."""

import os

from addonmanager_freecad_interface import translate
from PySideWrapper import QtCore, QtGui, QtWidgets


class WidgetGlobalButtonBar(QtWidgets.QWidget):
    """A QWidget-derived class for displaying a set of buttons that affect the Addon Manager as a
    whole (rather than a specific Addon)."""

    def __init__(self, parent: QtWidgets.QWidget = None):
        super().__init__(parent)
        self.horizontal_layout = None
        self.update_all_addons = None
        self.check_for_updates = None
        self.python_dependencies = None
        self.addons_folder = None
        self.close = None
        self._update_ui()
        self.retranslateUi(None)
        self._set_icons()

    def _update_ui(self):
        self.horizontal_layout = QtWidgets.QHBoxLayout()
        self.update_all_addons = QtWidgets.QPushButton(self)
        self.check_for_updates = QtWidgets.QPushButton(self)
        self.close = QtWidgets.QPushButton(self)

        self.gear_tools = QtWidgets.QToolButton()
        icon_path = os.path.join(os.path.dirname(__file__), "..", "Resources", "icons")
        self.gear_tools.setIcon(QtGui.QIcon(os.path.join(icon_path, "gear.svg")))
        self.gear_tools.setToolButtonStyle(QtCore.Qt.ToolButtonIconOnly)
        self.gear_tools.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.gear_tools.setStyleSheet("QToolButton::menu-indicator { image: none; }")

        if hasattr(QtGui, "QAction"):
            # Qt6
            self.addons_folder = QtGui.QAction()
            self.python_dependencies = QtGui.QAction()
        else:
            # Qt5
            self.addons_folder = QtWidgets.QAction()
            self.python_dependencies = QtWidgets.QAction()
        self.gear_menu = QtWidgets.QMenu(self.gear_tools)
        self.gear_menu.addAction(self.addons_folder)
        self.gear_menu.addAction(self.python_dependencies)

        self.gear_tools.setMenu(self.gear_menu)

        self.horizontal_layout.addWidget(self.update_all_addons)
        self.horizontal_layout.addWidget(self.check_for_updates)
        self.horizontal_layout.addWidget(self.gear_tools)
        self.horizontal_layout.addStretch()
        self.horizontal_layout.addWidget(self.close)
        self.setLayout(self.horizontal_layout)

    def _set_icons(self):
        icon_path = os.path.join(os.path.dirname(__file__), "..", "Resources", "icons")
        self.update_all_addons.setIcon(QtGui.QIcon(os.path.join(icon_path, "button_valid.svg")))
        self.check_for_updates.setIcon(QtGui.QIcon(os.path.join(icon_path, "view-refresh.svg")))
        self.close.setIcon(
            QtGui.QIcon.fromTheme("close", QtGui.QIcon(os.path.join(icon_path, "process-stop.svg")))
        )

    def retranslateUi(self, _):
        self.update_all_addons.setText(translate("AddonsInstaller", "Update All Addons"))
        self.check_for_updates.setText(translate("AddonsInstaller", "Check for Updates"))
        self.python_dependencies.setText(translate("AddonsInstaller", "Open Python Dependencies"))
        self.addons_folder.setText(translate("AddonsFolder", "Open Addons Folder"))
        self.close.setText(translate("AddonsInstaller", "Close"))
        self.gear_tools.setText("Gear Tools…")  # This text should never be user-visible

    def set_number_of_available_updates(self, updates: int):
        if updates > 0:
            self.update_all_addons.setEnabled(True)
            self.update_all_addons.setText(
                translate("AddonsInstaller", "See %n Update(s)…", "", updates)
            )
        else:
            self.update_all_addons.setEnabled(False)
            self.update_all_addons.setText(translate("AddonsInstaller", "No updates available"))



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint transform"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


import FreeCAD

from femtools import geomtools


def get_analysis_types():
    return "all"  # write for all analysis types


def get_sets_name():
    return "constraints_transform_node_sets"


def get_constraint_title():
    return "Transform Constraints"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return ""


def get_after_write_constraint():
    return ""


def write_meshdata_constraint(f, femobj, trans_obj, ccxwriter):
    if trans_obj.TransformType == "Rectangular":
        f.write(f"*NSET,NSET=Rect{trans_obj.Name}\n")
    elif trans_obj.TransformType == "Cylindrical":
        f.write(f"*NSET,NSET=Cylin{trans_obj.Name}\n")
    for n in femobj["Nodes"]:
        f.write(f"{n},\n")


def write_constraint(f, femobj, trans_obj, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module

    trans_name = ""
    trans_type = ""
    if trans_obj.TransformType == "Rectangular":
        trans_name = "Rect"
        trans_type = "R"
        x = trans_obj.Rotation * FreeCAD.Vector(1, 0, 0)
        y = trans_obj.Rotation * FreeCAD.Vector(0, 1, 0)
        coords = list(x) + list(y)
    elif trans_obj.TransformType == "Cylindrical":
        trans_name = "Cylin"
        trans_type = "C"
        base = trans_obj.BasePoint
        axis = trans_obj.Axis
        coords = list(base) + list(base + axis)
    f.write(
        "*TRANSFORM, NSET={}{}, TYPE={}\n".format(
            trans_name,
            trans_obj.Name,
            trans_type,
        )
    )
    f.write(
        "{:.13G},{:.13G},{:.13G},{:.13G},{:.13G},{:.13G}\n".format(
            coords[0],
            coords[1],
            coords[2],
            coords[3],
            coords[4],
            coords[5],
        )
    )



# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest
import unittest.mock
import pathlib
from tempfile import TemporaryDirectory
from PySide import QtCore, QtGui
from CAMTests.PathTestUtils import PathTestWithAssets
from Path.Tool.assets import DummyAssetSerializer
from Path.Tool.shape.models.icon import (
    ToolBitShapeIcon,
    ToolBitShapeSvgIcon,
    ToolBitShapePngIcon,
)


class TestToolBitShapeIconBase(PathTestWithAssets):
    """Base class for ToolBitShapeIcon tests."""

    ICON_CLASS = ToolBitShapeIcon

    def setUp(self):
        super().setUp()
        # Ensure a QApplication exists for QPixmap tests
        self.app = QtGui.QApplication.instance()

        # Create a test shape and a test SVG icon.
        self.test_shape = self.assets.get("toolbitshape://ballend")
        self.test_svg = self.test_shape.icon
        assert self.test_svg is not None
        self.icon = self.ICON_CLASS("test_icon_base", b"")

    def tearDown(self):
        self.app = None
        return super().tearDown()

    def test_create_instance(self):
        # Test basic instance creation
        icon_id = "test_icon_123.dat"
        icon = self.ICON_CLASS(icon_id, b"")
        self.assertEqual(icon.get_id(), icon_id)
        self.assertEqual(icon.data, b"")
        self.assertIsInstance(icon.abbreviations, dict)

    def test_to_bytes(self):
        # Test serializing to bytes
        icon_id = "test_to_bytes.bin"
        icon_data = b"some_binary_data"
        icon = ToolBitShapeIcon(icon_id, icon_data)
        self.assertEqual(icon.to_bytes(DummyAssetSerializer), icon_data)

    def test_get_size_in_bytes(self):
        # Test getting icon data length
        icon_with_data = ToolBitShapeIcon("with_data.bin", b"abc")
        self.assertEqual(icon_with_data.get_size_in_bytes(), 3)

        icon_no_data = ToolBitShapeIcon("no_data.dat", b"")
        self.assertEqual(icon_no_data.get_size_in_bytes(), 0)

    @unittest.mock.patch("Path.Tool.shape.util.create_thumbnail_from_data")
    def test_from_shape_data_success(self, mock_create_thumbnail):
        # Test creating instance from shape data - success case
        shape_id = "test_shape"
        thumbnail_data = b"png thumbnail data"
        mock_create_thumbnail.return_value = thumbnail_data
        icon = ToolBitShapeIcon.from_shape_data(self.test_svg.data, shape_id)

        mock_create_thumbnail.assert_called_once_with(self.test_svg.data)
        self.assertIsNotNone(icon)
        self.assertIsInstance(icon, ToolBitShapePngIcon)
        self.assertEqual(icon.get_id(), shape_id)
        self.assertEqual(icon.data, thumbnail_data)
        self.assertEqual(icon.abbreviations, {})

    @unittest.mock.patch("Path.Tool.shape.util.create_thumbnail_from_data")
    def test_from_shape_data_failure(self, mock_create_thumbnail):
        # Test creating instance from shape data - failure case
        shape_id = "test_shape"
        mock_create_thumbnail.return_value = None
        icon_failed = ToolBitShapeIcon.from_shape_data(self.test_svg.data, shape_id)

        mock_create_thumbnail.assert_called_once_with(self.test_svg.data)
        self.assertIsNone(icon_failed)

    def test_get_png(self):
        if not self.app:
            self.skipTest("QApplication not available, skipping test_get_png")
        if type(self) is TestToolBitShapeIconBase:
            self.skipTest("Skipping test on abstract base class")
        # Test getting PNG data from the icon
        icon_size = QtCore.QSize(16, 16)
        png_data = self.icon.get_png(icon_size)
        self.assertIsInstance(png_data, bytes)
        self.assertTrue(len(png_data) > 0)

    def test_get_qpixmap(self):
        if not self.app:
            self.skipTest("QApplication not available, skipping test_get_qpixmap")
        if type(self) is TestToolBitShapeIconBase:
            self.skipTest("Skipping test on abstract base class")
        # Test getting QPixmap from the icon
        icon_size = QtCore.QSize(31, 32)
        pixmap = self.icon.get_qpixmap(icon_size)
        self.assertIsInstance(pixmap, QtGui.QPixmap)
        self.assertFalse(pixmap.isNull())
        self.assertEqual(pixmap.size().width(), 31)
        self.assertEqual(pixmap.size().height(), 32)


class TestToolBitShapeSvgIcon(TestToolBitShapeIconBase):
    """Tests specifically for ToolBitShapeSvgIcon."""

    ICON_CLASS = ToolBitShapeSvgIcon

    def setUp(self):
        super().setUp()
        self.icon = ToolBitShapeSvgIcon("test_icon_svg", self.test_svg.data)

    def test_from_bytes_svg(self):
        # Test creating instance from bytes with SVG
        icon_svg = ToolBitShapeSvgIcon.from_bytes(
            self.test_svg.data, "test_from_bytes.svg", {}, DummyAssetSerializer
        )
        self.assertEqual(icon_svg.get_id(), "test_from_bytes.svg")
        self.assertEqual(icon_svg.data, self.test_svg.data)
        self.assertIsInstance(icon_svg.abbreviations, dict)

    def test_round_trip_serialization_svg(self):
        # Test serialization and deserialization round trip for SVG
        svg_id = "round_trip_svg.svg"
        icon_svg = ToolBitShapeSvgIcon(svg_id, self.test_svg.data)
        serialized_svg = icon_svg.to_bytes(DummyAssetSerializer)
        deserialized_svg = ToolBitShapeSvgIcon.from_bytes(
            serialized_svg, svg_id, {}, DummyAssetSerializer
        )
        self.assertEqual(deserialized_svg.get_id(), svg_id)
        self.assertEqual(deserialized_svg.data, self.test_svg.data)
        # Abbreviations are extracted on access, so we don't check the dict directly
        self.assertIsInstance(deserialized_svg.abbreviations, dict)

    def test_from_file_svg(self):
        # We cannot use NamedTemporaryFile on Windows, because there
        # we may not have permission to read the tempfile while it is
        # still open.
        # So we use TemporaryDirectory instead, to ensure cleanup while
        # still having a the temporary file inside it.
        with TemporaryDirectory() as thedir:
            tempfile = pathlib.Path(thedir, "test.svg")
            tempfile.write_bytes(self.test_svg.data)

            icon_id = "dummy_icon"
            icon = ToolBitShapeIcon.from_file(tempfile, icon_id)
            self.assertIsInstance(icon, ToolBitShapeSvgIcon)
            self.assertEqual(icon.get_id(), icon_id)
            self.assertEqual(icon.data, self.test_svg.data)
            self.assertIsInstance(icon.abbreviations, dict)

    def test_abbreviations_cached_property_svg(self):
        # Test abbreviations property and caching for SVG
        icon_svg = ToolBitShapeSvgIcon("cached_abbr.svg", self.test_svg.data)

        # Accessing the property should call the static method
        with unittest.mock.patch.object(
            ToolBitShapeSvgIcon, "get_abbreviations_from_svg"
        ) as mock_get_abbr:
            mock_get_abbr.return_value = {"param1": "A1"}
            abbr1 = icon_svg.abbreviations
            abbr2 = icon_svg.abbreviations
            mock_get_abbr.assert_called_once_with(self.test_svg.data)
            self.assertEqual(abbr1, {"param1": "A1"})
            self.assertEqual(abbr2, {"param1": "A1"})

    def test_get_abbr_svg(self):
        # Test getting abbreviations for SVG
        icon_data = self.test_svg.data
        icon = ToolBitShapeSvgIcon("abbr_test.svg", icon_data)
        # Assuming the test_svg data has 'diameter' and 'length' ids
        self.assertIsNotNone(icon.get_abbr("Diameter"))
        self.assertIsNotNone(icon.get_abbr("Length"))
        self.assertIsNone(icon.get_abbr("NonExistent"))

    def test_get_abbreviations_from_svg_static(self):
        # Test static method get_abbreviations_from_svg
        svg_content = self.test_svg.data
        abbr = ToolBitShapeSvgIcon.get_abbreviations_from_svg(svg_content)
        # Assuming the test_svg data has 'diameter' and 'length' ids
        self.assertIn("Diameter", abbr)
        self.assertIn("Length", abbr)

        # Test with invalid SVG
        invalid_svg = b"<svg><text id='param1'>A1</text>"  # Missing closing tag
        abbr_invalid = ToolBitShapeSvgIcon.get_abbreviations_from_svg(invalid_svg)
        self.assertEqual(abbr_invalid, {})

        # Test with no text elements
        no_text_svg = b'<svg xmlns="http://www.w3.org/2000/svg"><rect/></svg>'
        abbr_no_text = ToolBitShapeSvgIcon.get_abbreviations_from_svg(no_text_svg)
        self.assertEqual(abbr_no_text, {})


class TestToolBitShapePngIcon(TestToolBitShapeIconBase):
    """Tests specifically for ToolBitShapePngIcon."""

    ICON_CLASS = ToolBitShapePngIcon

    def setUp(self):
        super().setUp()
        self.png_data = b"\x89PNG\r\n\x1a\n"  # Basic PNG signature
        self.icon = ToolBitShapePngIcon("test_icon_png", self.png_data)

    def test_from_bytes_png(self):
        # Test creating instance from bytes with PNG
        icon_png = ToolBitShapePngIcon.from_bytes(
            self.png_data, "test_from_bytes.png", {}, DummyAssetSerializer
        )
        self.assertEqual(icon_png.get_id(), "test_from_bytes.png")
        self.assertEqual(icon_png.data, self.png_data)
        self.assertEqual(icon_png.abbreviations, {})  # No abbreviations for PNG

    def test_round_trip_serialization_png(self):
        # Test serialization and deserialization round trip for PNG
        png_id = "round_trip_png.png"
        serialized_png = self.icon.to_bytes(DummyAssetSerializer)
        deserialized_png = ToolBitShapePngIcon.from_bytes(
            serialized_png, png_id, {}, DummyAssetSerializer
        )
        self.assertEqual(deserialized_png.get_id(), png_id)
        self.assertEqual(deserialized_png.data, self.png_data)
        self.assertEqual(deserialized_png.abbreviations, {})

    def test_from_file_png(self):
        png_data = b"\\x89PNG\\r\\n\\x1a\\n"
        # We cannot use NamedTemporaryFile on Windows, because there
        # we may not have permission to read the tempfile while it is
        # still open.
        # So we use TemporaryDirectory instead, to ensure cleanup while
        # still having a the temporary file inside it.
        with TemporaryDirectory() as thedir:
            tempfile = pathlib.Path(thedir, "test.png")
            tempfile.write_bytes(png_data)

            icon_id = "dummy_icon"
            icon = ToolBitShapeIcon.from_file(tempfile, icon_id)
            self.assertIsInstance(icon, ToolBitShapePngIcon)
            self.assertEqual(icon.get_id(), icon_id)
            self.assertEqual(icon.data, png_data)
            self.assertEqual(icon.abbreviations, {})

    def test_abbreviations_cached_property_png(self):
        # Test abbreviations property and caching for PNG
        self.assertEqual(self.icon.abbreviations, {})

    def test_get_abbr_png(self):
        # Test getting abbreviations for PNG
        self.assertIsNone(self.icon.get_abbr("Diameter"))

    def test_get_qpixmap(self):
        self.skipTest("Skipping test, have no test data")


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the BSpline object."""
## @package bspline
# \ingroup draftobjects
# \brief Provides the object code for the BSpline object.

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
from draftobjects.base import DraftObject
from draftutils import gui_utils
from draftutils import params


class BSpline(DraftObject):
    """The BSpline object"""

    def __init__(self, obj):
        super().__init__(obj, "BSpline")

        _tip = QT_TRANSLATE_NOOP("App::Property", "The points of the B-spline")
        obj.addProperty("App::PropertyVectorList", "Points", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "If the B-spline is closed or not")
        obj.addProperty("App::PropertyBool", "Closed", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Create a face if this B-spline is closed")
        obj.addProperty("App::PropertyBool", "MakeFace", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "The area of this object")
        obj.addProperty("App::PropertyArea", "Area", "Draft", _tip, locked=True)

        obj.MakeFace = params.get_param("MakeFaceMode")
        obj.Closed = False
        obj.Points = []
        self.assureProperties(obj)

    def onDocumentRestored(self, obj):
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(obj, vp_module="view_bspline", vp_class="ViewProviderBSpline")

    def assureProperties(self, obj):  # for Compatibility with older versions
        if not hasattr(obj, "Parameterization"):
            _tip = QT_TRANSLATE_NOOP("App::Property", "Parameterization factor")
            obj.addProperty("App::PropertyFloat", "Parameterization", "Draft", _tip, locked=True)
            obj.Parameterization = 1.0
            self.knotSeq = []

    def parameterization(self, pts, a, closed):
        """Computes a knot Sequence for a set of points.
        fac (0-1) : parameterization factor
        fac = 0 -> Uniform / fac=0.5 -> Centripetal / fac=1.0 -> Chord-Length
        """
        if closed:  # we need to add the first point as the end point
            pts.append(pts[0])
        params = [0]
        for i in range(1, len(pts)):
            p = pts[i].sub(pts[i - 1])
            pl = pow(p.Length, a)
            params.append(params[-1] + pl)
        return params

    def onChanged(self, fp, prop):
        self.props_changed_store(prop)

        if prop == "Parameterization":
            if fp.Parameterization < 0.0:
                fp.Parameterization = 0.0
            if fp.Parameterization > 1.0:
                fp.Parameterization = 1.0

    def execute(self, obj):
        if self.props_changed_placement_only() or not obj.Points:
            obj.positionBySupport()
            self.props_changed_clear()
            return

        import Part

        self.assureProperties(obj)

        self.knotSeq = self.parameterization(obj.Points, obj.Parameterization, obj.Closed)
        plm = obj.Placement
        if obj.Closed and (len(obj.Points) > 2):
            if obj.Points[0] == obj.Points[-1]:  # should not occur, but OCC will crash
                _err = QT_TRANSLATE_NOOP(
                    "Draft",
                    "_BSpline.createGeometry: "
                    "Closed with same first/last Point. Geometry not updated.",
                )
                App.Console.PrintError(_err + "\n")
                return
            spline = Part.BSplineCurve()
            spline.interpolate(obj.Points, PeriodicFlag=True, Parameters=self.knotSeq)
            # DNC: bug fix: convert to face if closed
            shape = Part.Wire(spline.toShape())
            # Creating a face from a closed spline cannot be expected to always work
            # Usually, if the spline is not flat the call of Part.Face() fails
            try:
                if hasattr(obj, "MakeFace"):
                    if obj.MakeFace:
                        shape = Part.Face(shape)
                else:
                    shape = Part.Face(shape)
            except Part.OCCError:
                pass
            obj.Shape = shape
            if hasattr(obj, "Area") and hasattr(shape, "Area"):
                obj.Area = shape.Area
        else:
            spline = Part.BSplineCurve()
            spline.interpolate(obj.Points, PeriodicFlag=False, Parameters=self.knotSeq)
            shape = spline.toShape()
            obj.Shape = shape
            if hasattr(obj, "Area") and hasattr(shape, "Area"):
                obj.Area = shape.Area
        obj.Placement = plm
        obj.positionBySupport()
        self.props_changed_clear()


# Alias for compatibility with v0.18 and earlier
_BSpline = BSpline

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""BIM nudge commands"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


class BIM_Nudge:
    # base class for the different nudge commands

    def getNudgeValue(self, mode):
        "mode can be dist, up, down, left, right. dist returns a float in mm, other modes return a 3D vector"

        from PySide import QtGui
        import WorkingPlane

        mw = FreeCADGui.getMainWindow()
        if mw:
            st = mw.statusBar()
            statuswidget = st.findChild(QtGui.QToolBar, "BIMStatusWidget")
            if statuswidget:
                nudgeValue = statuswidget.nudge.text().replace("&", "")
                dist = 0
                if "auto" in nudgeValue.lower():
                    unit = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Units").GetInt(
                        "UserSchema", 0
                    )
                    if unit in [2, 3, 5, 7]:
                        scale = [1.5875, 3.175, 6.35, 25.4, 152.4, 304.8]
                    else:
                        scale = [1, 5, 10, 50, 100, 500]
                    viewsize = (
                        FreeCADGui.ActiveDocument.ActiveView.getCameraNode()
                        .getViewVolume()
                        .getWidth()
                    )
                    if viewsize < 250:
                        dist = scale[0]
                    elif viewsize < 750:
                        dist = scale[1]
                    elif viewsize < 4500:
                        dist = scale[2]
                    elif viewsize < 8000:
                        dist = scale[3]
                    elif viewsize < 25000:
                        dist = scale[4]
                    else:
                        dist = scale[5]
                    # u = FreeCAD.Units.Quantity(dist,FreeCAD.Units.Length).UserString
                    statuswidget.nudge.setText(translate("BIM", "Auto"))
                else:
                    try:
                        dist = FreeCAD.Units.Quantity(nudgeValue)
                    except ValueError:
                        try:
                            dist = float(nudgeValue)
                        except ValueError:
                            return None
                    else:
                        dist = dist.Value
                if not dist:
                    return None
                if mode == "dist":
                    return dist
                wp = WorkingPlane.get_working_plane()
                if mode == "up":
                    return FreeCAD.Vector(wp.v).multiply(dist)
                if mode == "down":
                    return FreeCAD.Vector(wp.v).negative().multiply(dist)
                if mode == "right":
                    return FreeCAD.Vector(wp.u).multiply(dist)
                if mode == "left":
                    return FreeCAD.Vector(wp.u).negative().multiply(dist)
        return None

    def toStr(self, objs):
        "builds a string which is a list of objects"

        return "[" + ",".join(["FreeCAD.ActiveDocument." + obj.Name for obj in objs]) + "]"

    def getCenter(self, objs):
        "returns the center point of a group of objects"

        bb = None
        for obj in objs:
            if hasattr(obj, "Shape") and hasattr(obj.Shape, "BoundBox"):
                if not bb:
                    bb = obj.Shape.BoundBox
                else:
                    bb.add(obj.Shape.BoundBox)
        if bb:
            return bb.Center
        else:
            return None


class BIM_Nudge_Switch(BIM_Nudge):

    def GetResources(self):
        return {
            "MenuText": QT_TRANSLATE_NOOP("BIM_Nudge_Switch", "Nudge Switch"),
            "Accel": "Ctrl+/",
        }

    def Activated(self):
        from PySide import QtGui

        mw = FreeCADGui.getMainWindow()
        if mw:
            st = mw.statusBar()
            statuswidget = st.findChild(QtGui.QToolBar, "BIMStatusWidget")
            if statuswidget:
                nudgeValue = statuswidget.nudge.text()
                nudge = self.getNudgeValue("dist")
                if nudge:
                    u = FreeCAD.Units.Quantity(nudge, FreeCAD.Units.Length).UserString
                    if "auto" in nudgeValue.lower():
                        statuswidget.nudge.setText(u)
                    else:
                        statuswidget.nudge.setText(translate("BIM", "Auto"))


class BIM_Nudge_Up(BIM_Nudge):

    def GetResources(self):
        return {
            "MenuText": QT_TRANSLATE_NOOP("BIM_Nudge_Up", "Nudge Up"),
            "Accel": "Ctrl+Up",
        }

    def Activated(self):
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            nudge = self.getNudgeValue("up")
            if nudge:
                FreeCADGui.addModule("Draft")
                FreeCADGui.doCommand(
                    "Draft.move(" + self.toStr(sel) + ",FreeCAD." + str(nudge) + ")"
                )
                FreeCADGui.doCommand("FreeCAD.ActiveDocument.recompute()")


class BIM_Nudge_Down(BIM_Nudge):

    def GetResources(self):
        return {
            "MenuText": QT_TRANSLATE_NOOP("BIM_Nudge_Down", "Nudge Down"),
            "Accel": "Ctrl+Down",
        }

    def Activated(self):
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            nudge = self.getNudgeValue("down")
            if nudge:
                FreeCADGui.addModule("Draft")
                FreeCADGui.doCommand(
                    "Draft.move(" + self.toStr(sel) + ",FreeCAD." + str(nudge) + ")"
                )
                FreeCADGui.doCommand("FreeCAD.ActiveDocument.recompute()")


class BIM_Nudge_Left(BIM_Nudge):

    def GetResources(self):
        return {
            "MenuText": QT_TRANSLATE_NOOP("BIM_Nudge_Left", "Nudge Left"),
            "Accel": "Ctrl+Left",
        }

    def Activated(self):
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            nudge = self.getNudgeValue("left")
            if nudge:
                FreeCADGui.addModule("Draft")
                FreeCADGui.doCommand(
                    "Draft.move(" + self.toStr(sel) + ",FreeCAD." + str(nudge) + ")"
                )
                FreeCADGui.doCommand("FreeCAD.ActiveDocument.recompute()")


class BIM_Nudge_Right(BIM_Nudge):

    def GetResources(self):
        return {
            "MenuText": QT_TRANSLATE_NOOP("BIM_Nudge_Right", "Nudge Right"),
            "Accel": "Ctrl+Right",
        }

    def Activated(self):
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            nudge = self.getNudgeValue("right")
            if nudge:
                FreeCADGui.addModule("Draft")
                FreeCADGui.doCommand(
                    "Draft.move(" + self.toStr(sel) + ",FreeCAD." + str(nudge) + ")"
                )
                FreeCADGui.doCommand("FreeCAD.ActiveDocument.recompute()")


class BIM_Nudge_Extend(BIM_Nudge):

    def GetResources(self):
        return {
            "MenuText": QT_TRANSLATE_NOOP("BIM_Nudge_Extend", "Nudge Extend"),
            "Accel": "Ctrl+PgUp",
        }

    def Activated(self):
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            nudge = self.getNudgeValue("dist")
            if nudge:
                for obj in sel:
                    if hasattr(obj, "Height"):
                        FreeCADGui.doCommand(
                            "FreeCAD.ActiveDocument."
                            + obj.Name
                            + ".Height="
                            + str(obj.Height.Value + nudge)
                        )
                        FreeCADGui.doCommand("FreeCAD.ActiveDocument.recompute()")


class BIM_Nudge_Shrink(BIM_Nudge):

    def GetResources(self):
        return {
            "MenuText": QT_TRANSLATE_NOOP("BIM_Nudge_Shrink", "Nudge Shrink"),
            "Accel": "Ctrl+PgDown",
        }

    def Activated(self):
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            nudge = self.getNudgeValue("dist")
            if nudge:
                for obj in sel:
                    if hasattr(obj, "Height"):
                        FreeCADGui.doCommand(
                            "FreeCAD.ActiveDocument."
                            + obj.Name
                            + ".Height="
                            + str(obj.Height.Value - nudge)
                        )
                        FreeCADGui.doCommand("FreeCAD.ActiveDocument.recompute()")


class BIM_Nudge_RotateLeft(BIM_Nudge):

    def GetResources(self):
        return {
            "MenuText": QT_TRANSLATE_NOOP("BIM_Nudge_RotateLeft", "Nudge Rotate Left"),
            "Accel": "Ctrl+,",
        }

    def Activated(self):

        import WorkingPlane

        sel = FreeCADGui.Selection.getSelection()
        if sel:
            center = self.getCenter(sel)
            if center:
                FreeCADGui.addModule("Draft")
                FreeCADGui.doCommand(
                    "Draft.rotate("
                    + self.toStr(sel)
                    + ",45,FreeCAD."
                    + str(center)
                    + ",FreeCAD."
                    + str(WorkingPlane.get_working_plane().axis)
                    + ")"
                )
                FreeCADGui.doCommand("FreeCAD.ActiveDocument.recompute()")


class BIM_Nudge_RotateRight(BIM_Nudge):

    def GetResources(self):
        return {
            "MenuText": QT_TRANSLATE_NOOP("BIM_Nudge_RotateRight", "Nudge Rotate Right"),
            "Accel": "Ctrl+.",
        }

    def Activated(self):

        import WorkingPlane

        sel = FreeCADGui.Selection.getSelection()
        if sel:
            center = self.getCenter(sel)
            if center:
                FreeCADGui.addModule("Draft")
                FreeCADGui.doCommand(
                    "Draft.rotate("
                    + self.toStr(sel)
                    + ",-45,FreeCAD."
                    + str(center)
                    + ",FreeCAD."
                    + str(WorkingPlane.get_working_plane().axis)
                    + ")"
                )
                FreeCADGui.doCommand("FreeCAD.ActiveDocument.recompute()")


FreeCADGui.addCommand("BIM_Nudge_Switch", BIM_Nudge_Switch())
FreeCADGui.addCommand("BIM_Nudge_Up", BIM_Nudge_Up())
FreeCADGui.addCommand("BIM_Nudge_Down", BIM_Nudge_Down())
FreeCADGui.addCommand("BIM_Nudge_Left", BIM_Nudge_Left())
FreeCADGui.addCommand("BIM_Nudge_Right", BIM_Nudge_Right())
FreeCADGui.addCommand("BIM_Nudge_Extend", BIM_Nudge_Extend())
FreeCADGui.addCommand("BIM_Nudge_Shrink", BIM_Nudge_Shrink())
FreeCADGui.addCommand("BIM_Nudge_RotateLeft", BIM_Nudge_RotateLeft())
FreeCADGui.addCommand("BIM_Nudge_RotateRight", BIM_Nudge_RotateRight())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the viewprovider code for the BSpline object.

At the moment this view provider subclasses the Wire view provider,
and behaves the same as it. In the future this could change
if another behavior is desired.
"""
## @package view_bspline
# \ingroup draftviewproviders
# \brief Provides the view provider code for BSpline objects.

## \addtogroup draftviewproviders
# @{
from draftviewproviders.view_wire import ViewProviderWire


class ViewProviderBSpline(ViewProviderWire):
    """The view provider for the BSpline object."""

    def __init__(self, vobj):
        super(ViewProviderBSpline, self).__init__(vobj)


# Alias for compatibility with v0.18 and earlier
_ViewProviderBSpline = ViewProviderBSpline

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import FreeCAD
import Path
from ...shape import ToolBitShapeRadius
from ..mixins import RotaryToolBitMixin, CuttingToolMixin
from .base import ToolBit


class ToolBitRadius(ToolBit, CuttingToolMixin, RotaryToolBitMixin):
    SHAPE_CLASS = ToolBitShapeRadius

    def __init__(self, shape: ToolBitShapeRadius, id: str | None = None):
        Path.Log.track(f"ToolBitRadius __init__ called with shape: {shape}, id: {id}")
        super().__init__(shape, id=id)
        CuttingToolMixin.__init__(self, self.obj)

    @property
    def summary(self) -> str:
        radius = self.get_property_str("CuttingRadius", "?", precision=3)
        flutes = self.get_property("Flutes")
        diameter = self.get_property_str("ShankDiameter", "?", precision=3)

        return FreeCAD.Qt.translate(
            "CAM", f"R{radius} radius mill, {diameter} shank, {flutes}-flute"
        )
